""" Contains a component that uses the UTSP to provide LoadProfileGenerator data. """

# clean

import datetime
import errno
import io
import json
import os
from ast import literal_eval
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union
import copy
import enum
import pandas as pd
from dataclasses_json import dataclass_json

from utspclient import client, datastructures, result_file_filters
from utspclient.helpers import lpg_helper
from utspclient.helpers.lpgpythonbindings import HouseCreationAndCalculationJob
from utspclient.datastructures import TimeSeriesRequest
from utspclient.client import calculate_multiple_requests
from utspclient.helpers.lpgdata import (
    ChargingStationSets,
    Households,
    HouseTypes,
    LoadTypes,
    TransportationDeviceSets,
    TravelRouteSets,
    EnergyIntensityType,
)
from utspclient.helpers.lpgpythonbindings import CalcOption, JsonReference
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiHelperClass, KpiTagEnumClass

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log, utils
from hisim.components.configuration import HouseholdWarmWaterDemandConfig, PhysicsConfig
from hisim.simulationparameters import SimulationParameters
from hisim.component import OpexCostDataClass
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum


class LpgDataAcquisitionMode(enum.Enum):

    """Set LPG Data Acquisition Mode."""

    USE_PREDEFINED_PROFILE = "use_predefined_profile"
    USE_UTSP = "use_utsp"


@dataclass_json
@dataclass
class UtspLpgConnectorConfig(cp.ConfigBase):

    """Config class for UtspLpgConnector. Contains LPG parameters and UTSP connection parameters."""

    building_name: str
    name: str
    data_acquisition_mode: LpgDataAcquisitionMode
    household: Union[JsonReference, List[JsonReference]]
    energy_intensity: EnergyIntensityType
    travel_route_set: JsonReference
    transportation_device_set: JsonReference
    charging_station_set: JsonReference
    profile_with_washing_machine_and_dishwasher: bool
    predictive_control: bool
    predictive: bool
    result_dir_path: str
    cache_dir_path: Optional[str] = None
    guid: str = ""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return UtspLpgConnector.get_full_classname()

    @classmethod
    def get_default_utsp_connector_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Creates a default configuration. Chooses default values for the LPG parameters."""

        config = UtspLpgConnectorConfig(
            building_name=building_name,
            name="UTSPConnector",
            data_acquisition_mode=LpgDataAcquisitionMode.USE_PREDEFINED_PROFILE,
            household=Households.CHR01_Couple_both_at_Work,
            result_dir_path=utils.HISIMPATH["utsp_results"],
            energy_intensity=EnergyIntensityType.EnergySaving,
            travel_route_set=TravelRouteSets.Travel_Route_Set_for_10km_Commuting_Distance,
            transportation_device_set=TransportationDeviceSets.Bus_and_one_30_km_h_Car,
            charging_station_set=ChargingStationSets.Charging_At_Home_with_11_kW,
            profile_with_washing_machine_and_dishwasher=True,
            predictive_control=False,
            predictive=False,
            cache_dir_path=None,
            guid="",
        )
        return config


class UtspLpgConnector(cp.Component):
    """Component that provides data from the LoadProfileGenerator.

    This component provides the heating generated, the electricity and water consumed
    by the residents. Furthermore, transportation and device flexibility data is stored
    in separate files, configurable via the config object.
    The data is retrieved from the UTSP, which executes the LoadProfileGenerator to simulate
    the specified household.
    """

    # Inputs
    WW_MassInput = "Warm Water Mass Input"  # kg/s
    WW_TemperatureInput = "Warm Water Temperature Input"  # °C

    # Outputs
    WW_MassOutput = "Mass Output"  # kg/s
    WW_TemperatureOutput = "Temperature Output"  # °C
    EnergyDischarged = "Energy Discharged"  # W
    DemandSatisfied = "Demand Satisfied"  # 0 or 1

    NumberByResidents = "NumberByResidents"
    HeatingByResidents = "HeatingByResidents"
    HeatingByDevices = "HeatingByDevices"
    ElectricityOutput = "ElectricityOutput"
    ElectricityEnergyOutput = "ElectricityEnergyOutput"
    WaterConsumption = "WaterConsumption"

    Electricity_Demand_Forecast_24h = "Electricity_Demand_Forecast_24h"

    # Similar components to connect to:
    # None
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: UtspLpgConnectorConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initializes the component and retrieves the LPG data."""
        self.utsp_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.build()
        # dummy value as long as there is no way to consider multiple households in one house
        self.scaling_factor_according_to_number_of_apartments: float = 1.0

        # Inputs - Not Mandatory
        self.ww_mass_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_MassInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.KG_PER_SEC,
            False,
        )
        self.ww_temperature_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_TemperatureInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            False,
        )

        self.number_of_residents_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.NumberByResidents,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description=f"here a description for LPG UTSP {self.NumberByResidents} will follow.",
        )
        self.heating_by_residents_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.HeatingByResidents,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for LPG UTSP {self.HeatingByResidents} will follow.",
        )
        self.heating_by_devices_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.HeatingByDevices,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description="Inner device heat gains, which heat the building (not intentionally)",
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
            output_description=f"here a description for LPG UTSP {self.ElectricityOutput} will follow.",
        )

        self.electricity_energy_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityEnergyOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            postprocessing_flag=[lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
            output_description=f"here a description for LPG UTSP {self.ElectricityEnergyOutput} will follow.",
        )

        self.water_consumption_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.WaterConsumption,
            lt.LoadTypes.WARM_WATER,
            lt.Units.LITER,
            output_description=f"here a description for LPG UTSP {self.WaterConsumption} will follow.",
        )

    def i_save_state(self) -> None:
        """Empty method as component has no state."""
        pass

    def i_restore_state(self) -> None:
        """Empty method as component has no state."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Gets called after the iterations are finished at each time step for potential debugging purposes."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Sets the current output values with data retrieved during initialization."""
        if self.ww_mass_input_channel.source_output is not None:
            # ww demand
            ww_temperature_demand = HouseholdWarmWaterDemandConfig.ww_temperature_demand

            # From Thermal Energy Storage
            ww_mass_input_per_sec = stsv.get_input_value(self.ww_mass_input_channel)  # kg/s
            # ww_mass_input = ww_mass_input_per_sec * self.seconds_per_timestep           # kg
            ww_mass_input: float = ww_mass_input_per_sec
            ww_temperature_input = stsv.get_input_value(self.ww_temperature_input_channel)  # °C

            # Information import
            freshwater_temperature = HouseholdWarmWaterDemandConfig.freshwater_temperature
            temperature_difference_cold = HouseholdWarmWaterDemandConfig.temperature_difference_cold
            energy_losses = 0
            specific_heat = 4180 / 3600

            ww_energy_demand = (
                specific_heat * self.water_consumption[timestep] * (ww_temperature_demand - freshwater_temperature)
            )

            if ww_energy_demand > 0 and (ww_mass_input == 0 and ww_temperature_input == 0):
                """first iteration --> random numbers"""
                ww_temperature_input = 40.45
                ww_mass_input = 9.3

            """
            Warm water is provided by the warmwater stoage.
            The household needs water at a certain temperature. To get the correct temperature the amount of water from
            the wws is regulated and is depending on the temperature provided by the wws. The backflowing water to wws
            is cooled down to the temperature of (freshwater+temperature_difference_cold) --> ww_temperature_output.
            """
            if ww_energy_demand > 0:
                # heating up the freshwater. The mass is consistent
                energy_discharged = ww_energy_demand + energy_losses
                ww_temperature_output: float = freshwater_temperature + temperature_difference_cold
                ww_mass_input = energy_discharged / (
                    PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
                    * (ww_temperature_input - ww_temperature_output)
                )
            else:
                ww_temperature_output = ww_temperature_input
                ww_mass_input = 0
                energy_discharged = 0

        stsv.set_output_value(self.number_of_residents_channel, self.number_of_residents[timestep])
        stsv.set_output_value(self.heating_by_residents_channel, self.heating_by_residents[timestep])
        stsv.set_output_value(self.heating_by_devices_channel, self.heating_by_devices[timestep])
        stsv.set_output_value(self.electricity_output_channel, self.electricity_consumption[timestep])
        stsv.set_output_value(
            self.electricity_energy_output_channel,
            self.electricity_consumption[timestep] * self.my_simulation_parameters.seconds_per_timestep / 3600,
        )
        stsv.set_output_value(self.water_consumption_channel, self.water_consumption[timestep])

        if self.config.predictive_control:
            last_forecast_timestep = int(timestep + 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep)
            if last_forecast_timestep > len(self.electricity_consumption):
                last_forecast_timestep = len(self.electricity_consumption)
            demandforecast = self.electricity_consumption[timestep:last_forecast_timestep]
            self.simulation_repository.set_entry(self.Electricity_Demand_Forecast_24h, demandforecast)

    def get_resolution(self) -> str:
        """Gets the temporal resolution of the simulation as a string in the format hh:mm:ss.

        :return: resolution of the simulation
        :rtype: str
        """
        seconds = self.my_simulation_parameters.seconds_per_timestep
        resolution = datetime.timedelta(seconds=seconds)
        return str(resolution)

    def get_profiles_from_utsp(self, lpg_households: Union[JsonReference, List[JsonReference]], guid: str) -> Tuple[
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
        Union[str, List],
    ]:
        """Requests the required load profiles from a UTSP server. Returns raw, unparsed result file contents.

        :return: a tuple of all result file contents (electricity, warm water, high bodily activity and low bodily activity),
                 and a list of filenames of all additionally saved files
        """
        # Create an LPG configuration and set the simulation parameters
        start_date = self.my_simulation_parameters.start_date.strftime("%Y-%m-%d")
        # Unlike HiSim the LPG includes the specified end day in the simulation --> subtract one day
        last_day = self.my_simulation_parameters.end_date - datetime.timedelta(days=1)
        end_date = last_day.strftime("%Y-%m-%d")

        # choose if one lpg request should be made or several in parallel
        if isinstance(lpg_households, JsonReference):
            simulation_config = self.prepare_lpg_simulation_config_for_utsp_request(
                start_date=start_date,
                end_date=end_date,
                household=lpg_households,
            )

            (
                electricity_file,
                warm_water_file,
                inner_device_heat_gains_file,
                high_activity_file,
                low_activity_file,
                flexibility_file,
                car_states_file,
                car_locations_file,
                driving_distances_file,
            ) = self.calculate_one_lpg_request(
                simulation_config=simulation_config,
                guid=guid,
            )

        elif isinstance(lpg_households, List):
            simulation_configs = []
            for household in lpg_households:
                simulation_config = self.prepare_lpg_simulation_config_for_utsp_request(
                    start_date=start_date, end_date=end_date, household=household
                )
                simulation_configs.append(simulation_config)

            (
                electricity_file,
                warm_water_file,
                inner_device_heat_gains_file,
                high_activity_file,
                low_activity_file,
                flexibility_file,
                car_states_file,
                car_locations_file,
                driving_distances_file,
            ) = self.calculate_multiple_lpg_requests(  # type: ignore
                url=self.utsp_url,
                api_key=self.utsp_api_key,
                lpg_configs=simulation_configs,
                guid=guid,
            )

        else:
            raise TypeError(
                f"Type of lpg_households {type(lpg_households)} is invalid. It should be a JSONReference or a list of JSONReference."
            )

        return (
            electricity_file,
            warm_water_file,
            inner_device_heat_gains_file,
            high_activity_file,
            low_activity_file,
            flexibility_file,
            car_states_file,
            car_locations_file,
            driving_distances_file,
        )

    def get_profiles_from_predefined_profile(
        self,
    ) -> Tuple[str, str, str, str, str]:
        """Get the loadprofiles for a specific predefined profile from hisim/inputs/loadprofiles."""
        predefined_profile_filepaths = utils.HISIMPATH["occupancy"]["CHR01 Couple both at Work"]
        # get first bodily activity files
        bodily_activity_filepaths = predefined_profile_filepaths["number_of_residents"]

        high_activity_file = bodily_activity_filepaths[0]
        low_activity_file = bodily_activity_filepaths[1]
        # get other files
        if self.utsp_config.profile_with_washing_machine_and_dishwasher:
            electricity_file = predefined_profile_filepaths["electricity_consumption"]
        else:
            electricity_file = predefined_profile_filepaths[
                "electricity_consumption_without_washing_machine_and_dishwasher"
            ]

        warm_water_file = predefined_profile_filepaths["water_consumption"]
        inner_device_heat_gains_file = predefined_profile_filepaths["heating_by_devices"]
        # when using predefined profile there are no saved files concerning flexibility or car data
        # TODO: implement flexibility and car files?
        return (
            electricity_file,
            warm_water_file,
            inner_device_heat_gains_file,
            high_activity_file,
            low_activity_file,
        )

    def save_result_file(self, name: str, content: str) -> str:
        """Saves a result file in the folder specified in the config object.

        :param name: the name for the file
        :type name: str
        :param content: the content that will be written into the file
        :type content: str
        :return: path of the file that was saved
        :rtype: str
        """
        try:
            filepath = os.path.join(self.utsp_config.result_dir_path, name)
        except Exception as exc:
            raise NameError(
                f"Could not create a filepath from config result_path {self.utsp_config.result_dir_path} and name {name}."
            ) from exc

        directory = os.path.dirname(filepath)
        # Create the directory if it does not exist
        try:
            os.makedirs(directory)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(directory):
                pass
            else:
                raise
        # Create the result file
        with open(filepath, "w", encoding="utf-8") as result_file:
            result_file.write(content)

        return filepath

    def build(self):
        """Retrieves and preprocesses all data for this component."""

        # check if file exists and get cache_filepath and put in list
        (
            self.list_of_file_exists_and_cache_files,
            list_of_unique_household_configs,
        ) = self.get_list_of_file_exists_bools_and_cache_file_paths(cache_dir_path=self.utsp_config.cache_dir_path)

        # go through list of file_exists and cache_filepaths and get caches if possible,
        # otherwise send request to UTSP
        cache_complete = False
        value_dict: Dict = {
            "electricity_consumption": [],
            "water_consumption": [],
            "heating_by_devices": [],
            "heating_by_residents": [],
            "number_of_residents": [],
        }
        self.car_data_dict: Dict = {
            "car_states": [],
            "car_locations": [],
            "driving_distances": [],
        }
        self.flexibility_data_dict: Dict = {"flexibility": []}
        # iterate over all unique utsp configs and either take cache results or calculate for each household and sum up later
        for list_index, list_item in enumerate(self.list_of_file_exists_and_cache_files):
            file_exists = list_item[0]
            cache_filepath = list_item[1]
            print("lpg cache file", cache_filepath)

            # a cache file exists
            if file_exists:
                with open(cache_filepath, "r", encoding="utf-8") as file:
                    cache_content: Dict = json.load(file)
                    cache_complete = True

                # check if cache content has correct format, otherwise delete cache (because it is from older version) and run new request
                if (
                    not all(isinstance(values, str) for values in cache_content.values())
                    and "saved_files" in cache_content.keys()
                ):
                    log.information(
                        "An older LPG cache version was found but it's not usuable anymore. Therefore it will be deleted."
                    )
                    os.remove(cache_filepath)
                    cache_complete = False

            if cache_complete:
                log.information("LPG data taken from cache. ")
                for cache_key in cache_content.keys():
                    # get dataframe from cache content
                    cached_data = io.StringIO(cache_content[cache_key])
                    dataframe = pd.read_csv(cached_data, sep=",", decimal=".", encoding="cp1252", index_col=0)

                    if cache_key == "data":

                        number_of_residents = dataframe["number_of_residents"].tolist()
                        heating_by_residents = dataframe["heating_by_residents"].tolist()
                        electricity_consumption = dataframe["electricity_consumption"].tolist()
                        water_consumption = dataframe["water_consumption"].tolist()
                        heating_by_devices = dataframe["heating_by_devices"].to_list()

                        number_of_residents = dataframe["number_of_residents"].tolist()

                        # write lists to dict
                        value_dict["electricity_consumption"].append(electricity_consumption)
                        value_dict["heating_by_devices"].append(heating_by_devices)
                        value_dict["heating_by_residents"].append(heating_by_residents)
                        value_dict["water_consumption"].append(water_consumption)
                        value_dict["number_of_residents"].append(number_of_residents)

                        # sum over all household profiles
                        (
                            self.electricity_consumption,
                            self.heating_by_residents,
                            self.water_consumption,
                            self.heating_by_devices,
                            self.number_of_residents,
                        ) = self.get_result_lists_by_summing_over_value_dict(value_dict=value_dict)

                        self.max_hot_water_demand = max(self.water_consumption)

                    elif cache_key == "car_data":
                        # process car data from cache
                        for key, dict_values in self.car_data_dict.items():
                            # transform dataframe column to dict, then get original types of dict values and lastly append to
                            # car_data_dict
                            transformed_data_dict = self.transform_dict_values(dataframe[key].to_dict())
                            dict_values.append(transformed_data_dict)

                    elif cache_key == "flexibility_data":
                        # process flexibility data from cache
                        # transform dataframe column to dict, then get original types of dict values and lastly append to
                        # car_and_flexibility_dict
                        transformed_data_dict = self.transform_dict_values(dataframe["flexibility"].to_dict())
                        self.flexibility_data_dict["flexibility"].append(transformed_data_dict)

                    else:
                        raise KeyError(f"The cache content key {cache_key} could not be recognized.")

            if not cache_complete or file_exists is False:
                log.information(
                    "LPG data cannot be taken from cache. It will be taken from UTSP or from predefined profile."
                )
                # if taking results from cache not possible, check lpg data acquition mode
                if self.utsp_config.data_acquisition_mode == LpgDataAcquisitionMode.USE_UTSP:
                    # try to get utsp url and api from .env if possible
                    try:
                        self.utsp_url = utils.get_environment_variable("UTSP_URL")
                        self.utsp_api_key = utils.get_environment_variable("UTSP_API_KEY")

                    except Exception:
                        log.warning(
                            "You chose USE_UTSP as data_acquition_mode but it is not possible to read the url and api_key from the .env file."
                            "Please check if this file is present in your system."
                            "Otherwise the predefined LPG profile in hisim/inputs/loadprofiles will be used."
                        )
                        self.utsp_config.data_acquisition_mode = LpgDataAcquisitionMode.USE_PREDEFINED_PROFILE

                if self.utsp_config.data_acquisition_mode == LpgDataAcquisitionMode.USE_UTSP:
                    log.information(f"LPG data acquisition mode: {self.utsp_config.data_acquisition_mode}")
                    new_unique_config = list_of_unique_household_configs[list_index]

                    (
                        electricity_file,
                        warm_water_file,
                        inner_device_heat_gains_file,
                        high_activity_file,
                        low_activity_file,
                        flexibility_file,
                        car_states_file,
                        car_locations_file,
                        driving_distances_file,
                    ) = self.get_profiles_from_utsp(
                        lpg_households=new_unique_config.household,
                        guid=new_unique_config.guid,
                    )

                    # only one result obtained
                    if isinstance(electricity_file, str):
                        log.information("One result obtained from lpg utsp connector.")
                        (
                            electricity_consumption,
                            heating_by_devices,
                            water_consumption,
                            heating_by_residents,
                            number_of_residents,
                        ) = self.load_result_files_and_transform_to_lists(
                            electricity=electricity_file,
                            warm_water=warm_water_file,
                            inner_device_heat_gains=inner_device_heat_gains_file,
                            high_activity=high_activity_file,
                            low_activity=low_activity_file,
                            data_acquisition_mode=self.utsp_config.data_acquisition_mode,
                        )
                        list_of_flexibility_and_car_files = [
                            flexibility_file,
                            car_states_file,
                            car_locations_file,
                            driving_distances_file,
                        ]
                        if all(isinstance(file, str) for file in list_of_flexibility_and_car_files):
                            list_of_flexibility_and_car_data = self.load_results_and_transform_string_to_data(
                                list_of_result_files=list_of_flexibility_and_car_files  # type: ignore
                            )
                        else:
                            raise TypeError(
                                f"Type of flexibility and car files should be str, but it's {[type(i) for i in list_of_flexibility_and_car_files]}"
                            )

                        # write lists to dict
                        value_dict["electricity_consumption"].append(electricity_consumption)
                        value_dict["heating_by_devices"].append(heating_by_devices)
                        value_dict["heating_by_residents"].append(heating_by_residents)
                        value_dict["water_consumption"].append(water_consumption)
                        value_dict["number_of_residents"].append(number_of_residents)
                        self.flexibility_data_dict["flexibility"].append(list_of_flexibility_and_car_data[0])
                        self.car_data_dict["car_states"].append(list_of_flexibility_and_car_data[1])
                        self.car_data_dict["car_locations"].append(list_of_flexibility_and_car_data[2])
                        self.car_data_dict["driving_distances"].append(list_of_flexibility_and_car_data[3])

                        # cache results for each household individually
                        self.cache_results(
                            cache_filepath=cache_filepath,
                            number_of_residents=number_of_residents,
                            electricity_consumption=electricity_consumption,
                            heating_by_residents=heating_by_residents,
                            water_consumption=water_consumption,
                            heating_by_devices=heating_by_devices,
                            flexibility=list_of_flexibility_and_car_data[0],
                            car_states=list_of_flexibility_and_car_data[1],
                            car_locations=list_of_flexibility_and_car_data[2],
                            driving_distances=list_of_flexibility_and_car_data[3],
                        )

                    # multiple results obtained (when multiple households in utsp_config given and the guid in the config is not "" but has a specific value)
                    elif isinstance(electricity_file, List):
                        log.information("Multiple results obtained from lpg utsp connector.")

                        for index, electricity in enumerate(electricity_file):
                            warm_water = warm_water_file[index]
                            inner_device_heat_gains = inner_device_heat_gains_file[index]
                            high_activity = high_activity_file[index]
                            low_activity = low_activity_file[index]
                            flexibility = flexibility_file[index]
                            car_states = car_states_file[index]
                            car_locations = car_locations_file[index]
                            driving_distances = driving_distances_file[index]

                            (
                                electricity_consumption,
                                heating_by_devices,
                                water_consumption,
                                heating_by_residents,
                                number_of_residents,
                            ) = self.load_result_files_and_transform_to_lists(
                                electricity=electricity,
                                warm_water=warm_water,
                                inner_device_heat_gains=inner_device_heat_gains,
                                high_activity=high_activity,
                                low_activity=low_activity,
                                data_acquisition_mode=self.utsp_config.data_acquisition_mode,
                            )
                            list_of_flexibility_and_car_data = self.load_results_and_transform_string_to_data(
                                list_of_result_files=[flexibility, car_states, car_locations, driving_distances]
                            )

                            # write lists to dict
                            value_dict["electricity_consumption"].append(electricity_consumption)
                            value_dict["heating_by_devices"].append(heating_by_devices)
                            value_dict["heating_by_residents"].append(heating_by_residents)
                            value_dict["water_consumption"].append(water_consumption)
                            value_dict["number_of_residents"].append(number_of_residents)
                            self.flexibility_data_dict["flexibility"].append(list_of_flexibility_and_car_data[0])
                            self.car_data_dict["car_states"].append(list_of_flexibility_and_car_data[1])
                            self.car_data_dict["car_locations"].append(list_of_flexibility_and_car_data[2])
                            self.car_data_dict["driving_distances"].append(list_of_flexibility_and_car_data[3])

                        # get sum of all household profiles
                        (
                            self.electricity_consumption,
                            self.heating_by_residents,
                            self.water_consumption,
                            self.heating_by_devices,
                            self.number_of_residents,
                        ) = self.get_result_lists_by_summing_over_value_dict(value_dict=value_dict)

                        self.max_hot_water_demand = max(self.water_consumption)

                        # cache for multiple results at a time
                        self.cache_results(
                            cache_filepath=cache_filepath,
                            number_of_residents=self.number_of_residents,
                            heating_by_residents=self.heating_by_residents,
                            water_consumption=self.water_consumption,
                            heating_by_devices=self.heating_by_devices,
                            electricity_consumption=self.electricity_consumption,
                            flexibility=list_of_flexibility_and_car_data[0],
                            car_states=list_of_flexibility_and_car_data[1],
                            car_locations=list_of_flexibility_and_car_data[2],
                            driving_distances=list_of_flexibility_and_car_data[3],
                        )
                        break

                    # get sum of all household profiles
                    (
                        self.electricity_consumption,
                        self.heating_by_residents,
                        self.water_consumption,
                        self.heating_by_devices,
                        self.number_of_residents,
                    ) = self.get_result_lists_by_summing_over_value_dict(value_dict=value_dict)

                    self.max_hot_water_demand = max(self.water_consumption)

                elif self.utsp_config.data_acquisition_mode == LpgDataAcquisitionMode.USE_PREDEFINED_PROFILE:
                    log.information(
                        f"LPG data acquisition mode: {self.utsp_config.data_acquisition_mode}. "
                        "This means the predefined_lpg_household_chr01 from hisim/inputs/loadprofiles/ is taken."
                    )

                    (
                        electricity_file,
                        warm_water_file,
                        inner_device_heat_gains_file,
                        high_activity_file,
                        low_activity_file,
                    ) = self.get_profiles_from_predefined_profile()

                    (
                        self.electricity_consumption,
                        self.heating_by_devices,
                        self.water_consumption,
                        self.heating_by_residents,
                        self.number_of_residents,
                    ) = self.load_result_files_and_transform_to_lists(
                        electricity=electricity_file,
                        warm_water=warm_water_file,
                        inner_device_heat_gains=inner_device_heat_gains_file,
                        high_activity=high_activity_file,
                        low_activity=low_activity_file,
                        data_acquisition_mode=self.utsp_config.data_acquisition_mode,
                    )

                    self.max_hot_water_demand = max(self.water_consumption)

                    # no caching if predefined profile is used

                    if self.utsp_config.predictive:
                        SingletonSimRepository().set_entry(
                            key=SingletonDictKeyEnum.HEATINGBYRESIDENTSYEARLYFORECAST,
                            entry=self.heating_by_residents,
                        )

    def get_result_lists_by_summing_over_value_dict(
        self, value_dict: Dict[Any, Any]
    ) -> Tuple[List, List, List, List, List]:
        """Get the result lists by summing over the value dict entries."""

        electricity_consumption = [sum(x) for x in zip(*value_dict["electricity_consumption"])]
        heating_by_residents = [sum(x) for x in zip(*value_dict["heating_by_residents"])]
        water_consumption = [sum(x) for x in zip(*value_dict["water_consumption"])]
        heating_by_devices = [sum(x) for x in zip(*value_dict["heating_by_devices"])]
        number_of_residents = [sum(x) for x in zip(*value_dict["number_of_residents"])]

        return (
            electricity_consumption,
            heating_by_residents,
            water_consumption,
            heating_by_devices,
            number_of_residents,
        )

    def write_to_report(self):
        """Adds a report entry for this component."""
        return self.utsp_config.get_string_dict()

    def get_list_of_file_exists_bools_and_cache_file_paths(self, cache_dir_path: Optional[str]) -> Tuple[List, List]:
        """Check if file exists and get cache_filepath and put in list."""

        list_of_file_exists_and_cache_files: List = []
        list_of_unique_household_configs: List = []

        # check if cache_dir_path was chosen, otherwise use default cache_dir_path
        if cache_dir_path is None:
            cache_dir_path = self.my_simulation_parameters.cache_dir_path

        # config household is list of jsonreferences and no other guid than default is given ("")
        # if the guid = "" and multiple households are given as a list, each household will be calculated and cached individually
        if isinstance(self.utsp_config.household, List) and self.utsp_config.guid == "":
            # get specific guid list in order to prevent duplicated requests
            guid_list = (
                self.vary_guids_for_lpg_utsp_requests_if_config_household_is_a_list_and_contains_duplicated_household_types()
            )

            for index, household in enumerate(self.utsp_config.household):
                # make new config object with only one household in order to find local cache in cache_dir_path
                new_config_object = copy.deepcopy(self.utsp_config)
                new_config_object.household = household
                new_config_object.guid = guid_list[index]

                # check if cache for utsp config exists and get or make cache filepath
                file_exists, cache_filepath = utils.get_cache_file(
                    component_key=self.component_name,
                    parameter_class=new_config_object,
                    my_simulation_parameters=self.my_simulation_parameters,
                    cache_dir_path=cache_dir_path,
                )
                list_of_file_exists_and_cache_files.append([file_exists, cache_filepath])
                list_of_unique_household_configs.append(new_config_object)

        # config household is one jsonreference
        else:
            file_exists, cache_filepath = utils.get_cache_file(
                component_key=self.component_name,
                parameter_class=self.utsp_config,
                my_simulation_parameters=self.my_simulation_parameters,
                cache_dir_path=cache_dir_path,
            )
            list_of_file_exists_and_cache_files.append([file_exists, cache_filepath])
            # ustp config is already unique because only 1 household in it
            list_of_unique_household_configs.append(self.utsp_config)

        return list_of_file_exists_and_cache_files, list_of_unique_household_configs

    def vary_guids_for_lpg_utsp_requests_if_config_household_is_a_list_and_contains_duplicated_household_types(
        self,
    ) -> List[str]:
        """In case the lpg_utsp_connector config is given a list of households, it will be checked if the list contains any duplicates.

        If so, for each duplicate the guid will be varied to make sure that each lpg request delivers a unique profile.
        """
        # check if household is list and return guid list
        if isinstance(self.utsp_config.household, List):
            copied_households = copy.deepcopy(self.utsp_config.household)
            guid_list = []

            for household in self.utsp_config.household:
                number_of_duplicated_households = copied_households.count(household)
                if number_of_duplicated_households == 1:
                    guid_list.append(str(1))
                elif number_of_duplicated_households > 1:
                    guid_list.append(str(number_of_duplicated_households))
                    copied_households.remove(household)

        return guid_list

    def calculate_one_lpg_request(
        self, simulation_config: HouseCreationAndCalculationJob, guid: str
    ) -> Tuple[str, str, str, str, str, str, str, str, str]:
        """Calculate one lpg request."""

        # define required results files
        (
            result_files,
            electricity,
            warm_water,
            inner_device_heat_gains,
            high_activity,
            low_activity,
            flexibility,
            car_states,
            car_locations,
            driving_distances,
        ) = self.define_required_result_files()

        # Prepare the time series request
        request = datastructures.TimeSeriesRequest(
            simulation_config.to_json(), "LPG", required_result_files=result_files, guid=guid  # type: ignore
        )

        log.information("Requesting LPG profiles from the UTSP for one household.")

        # Request the time series
        result = client.request_time_series_and_wait_for_delivery(
            self.utsp_url, request, self.utsp_api_key, timeout=100
        )

        # decode required result files
        electricity_file = result.data[electricity].decode()
        warm_water_file = result.data[warm_water].decode()
        inner_device_heat_gains_file = result.data[inner_device_heat_gains].decode()
        high_activity_file = result.data[high_activity].decode()
        low_activity_file = result.data[low_activity].decode()
        if flexibility in result.data.keys():
            flexibility_file = result.data[flexibility].decode()
        else:
            flexibility_file = ""

        for car_state in car_states.keys():
            if car_state in result.data:
                car_states_file = result.data[car_state].decode()
        for car_location in car_locations.keys():
            if car_location in result.data:
                car_locations_file = result.data[car_location].decode()
        for driving_distance in driving_distances.keys():
            if driving_distance in result.data:
                driving_distances_file = result.data[driving_distance].decode()

        return (
            electricity_file,
            warm_water_file,
            inner_device_heat_gains_file,
            high_activity_file,
            low_activity_file,
            flexibility_file,
            car_states_file,
            car_locations_file,
            driving_distances_file,
        )

    def calculate_multiple_lpg_requests(
        self,
        lpg_configs: List,
        url: str,
        api_key: str,
        guid: str,
        raise_exceptions: bool = True,
        result_files: Any = None,
    ) -> Tuple[List[str], List[str], List[str], List[str], List[str], List[str], List[str], List[str], List[str]]:
        """Sends multiple lpg requests for parallel calculation and collects their results."""

        (
            result_files,
            electricity,
            warm_water,
            inner_device_heat_gains,
            high_activity,
            low_activity,
            flexibility,
            car_states,
            car_locations,
            driving_distances,
        ) = self.define_required_result_files()

        # Create all request objects
        all_requests: List[TimeSeriesRequest] = [
            TimeSeriesRequest(
                config.to_json(),
                "LPG",
                required_result_files=result_files,
                guid=guid,
            )
            for config in lpg_configs
        ]

        log.information("Requesting LPG profiles from the UTSP for multiple household.")

        results = calculate_multiple_requests(
            url,
            all_requests,
            api_key,
            raise_exceptions,
        )

        # append all results in lists
        electricity_file: List = []
        warm_water_file: List = []
        inner_device_heat_gains_file: List = []
        high_activity_file: List = []
        low_activity_file: List = []
        flexibility_file: List = []
        car_states_file: List = []
        car_locations_file: List = []
        driving_distances_file: List = []

        for result in results:
            if isinstance(result, Exception):
                raise ValueError("result is an exception. Something went wrong during the utsp request.")

            electricity_file_one_result = result.data[electricity].decode()
            warm_water_file_one_result = result.data[warm_water].decode()
            inner_device_heat_gains_file_one_result = result.data[inner_device_heat_gains].decode()
            high_activity_file_one_result = result.data[high_activity].decode()
            low_activity_file_one_result = result.data[low_activity].decode()

            try:
                flexibility_file_one_result = result.data[flexibility].decode()
            except Exception:
                pass

            for car_state in car_states.keys():
                if car_state in result.data:
                    car_states_file_one_result = result.data[car_state].decode()

            for car_location in car_locations.keys():
                if car_location in result.data:
                    car_locations_file_one_result = result.data[car_location].decode()

            for driving_distance in driving_distances.keys():
                if driving_distance in result.data:
                    driving_distances_file_one_result = result.data[driving_distance].decode()

            # append to lists
            electricity_file.append(electricity_file_one_result)
            warm_water_file.append(warm_water_file_one_result)
            inner_device_heat_gains_file.append(inner_device_heat_gains_file_one_result)
            high_activity_file.append(high_activity_file_one_result)
            low_activity_file.append(low_activity_file_one_result)
            flexibility_file.append(flexibility_file_one_result)
            car_states_file.append(car_states_file_one_result)
            car_locations_file.append(car_locations_file_one_result)
            driving_distances_file.append(driving_distances_file_one_result)

        return (
            electricity_file,
            warm_water_file,
            inner_device_heat_gains_file,
            high_activity_file,
            low_activity_file,
            flexibility_file,
            car_states_file,
            car_locations_file,
            driving_distances_file,
        )

    def prepare_lpg_simulation_config_for_utsp_request(
        self, start_date: Any, end_date: Any, household: JsonReference
    ) -> HouseCreationAndCalculationJob:
        """Prepare lpg simulation config for the utsp request."""

        simulation_config = lpg_helper.create_basic_lpg_config(
            household,
            HouseTypes.HT23_No_Infrastructure_at_all,
            start_date,
            end_date,
            self.get_resolution(),
            travel_route_set=self.utsp_config.travel_route_set,
            transportation_device_set=self.utsp_config.transportation_device_set,
            charging_station_set=self.utsp_config.charging_station_set,
            calc_options=[
                CalcOption.HouseholdSumProfilesFromDetailedDats,
                CalcOption.HouseholdSumProfilesCsvNoFlex,
                CalcOption.BodilyActivityStatistics,
                CalcOption.TansportationDeviceJsons,
                CalcOption.FlexibilityEvents,
            ],
        )
        assert simulation_config.CalcSpec is not None

        # Enable simulation of transportation and flexible devices
        simulation_config.CalcSpec.EnableTransportation = True
        simulation_config.CalcSpec.EnableFlexibility = True

        return simulation_config

    def define_required_result_files(self):
        """Define required result files."""

        # Define required result files
        electricity = result_file_filters.LPGFilters.sum_hh1(
            LoadTypes.Electricity,
            no_flex=not self.utsp_config.profile_with_washing_machine_and_dishwasher,
        )
        warm_water = result_file_filters.LPGFilters.sum_hh1(LoadTypes.Warm_Water, no_flex=False)
        inner_device_heat_gains = result_file_filters.LPGFilters.sum_hh1(
            LoadTypes.Inner_Device_Heat_Gains, no_flex=False
        )
        high_activity = result_file_filters.LPGFilters.BodilyActivity.HIGH
        low_activity = result_file_filters.LPGFilters.BodilyActivity.LOW
        flexibility = result_file_filters.LPGFilters.FLEXIBILITY_EVENTS
        required_files = {
            f: datastructures.ResultFileRequirement.REQUIRED
            for f in [electricity, warm_water, inner_device_heat_gains, high_activity, low_activity]
        }
        optional_files = {flexibility: datastructures.ResultFileRequirement.OPTIONAL}
        # Define transportation result files
        car_states = result_file_filters.LPGFilters.all_car_states_optional()
        car_locations = result_file_filters.LPGFilters.all_car_locations_optional()
        driving_distances = result_file_filters.LPGFilters.all_driving_distances_optional()
        result_files: Dict[str, Optional[datastructures.ResultFileRequirement]] = {
            **required_files,
            **optional_files,
            **car_states,
            **car_locations,
            **driving_distances,
        }

        return (
            result_files,
            electricity,
            warm_water,
            inner_device_heat_gains,
            high_activity,
            low_activity,
            flexibility,
            car_states,
            car_locations,
            driving_distances,
        )

    def load_result_files_and_transform_to_lists(
        self,
        data_acquisition_mode: LpgDataAcquisitionMode,
        electricity: Any,
        warm_water: Any,
        inner_device_heat_gains: Any,
        high_activity: Any,
        low_activity: Any,
    ) -> tuple[List, List, List, List, List]:
        """Load result files and transform to lists."""

        ################################
        # Calculates heating generated by residents and loads number of residents
        # Heat power generated per resident in W
        # mode 1: awake
        # mode 2: sleeping
        gain_per_person = [150, 100]

        # load occupancy profile
        occupancy_profile = []
        bodily_activity_files = [high_activity, low_activity]
        for filecontent in bodily_activity_files:
            if data_acquisition_mode == LpgDataAcquisitionMode.USE_UTSP:
                json_filex = json.loads(filecontent)
            # this is used for files from predefined profile
            elif data_acquisition_mode == LpgDataAcquisitionMode.USE_PREDEFINED_PROFILE:
                with open(filecontent, encoding="utf-8") as json_file:
                    json_filex = json.load(json_file)
            else:
                raise ValueError("Could not recognize data_acquisition_mode.")

            occupancy_profile.append(json_filex)

        # see how long csv files from LPG are to check if averaging has to be done and calculate desired length
        simulation_time_span = self.my_simulation_parameters.end_date - self.my_simulation_parameters.start_date
        minutes_per_timestep = int(self.my_simulation_parameters.seconds_per_timestep / 60)
        steps_desired = int(
            simulation_time_span.days * 24 * (3600 / self.my_simulation_parameters.seconds_per_timestep)
        )
        steps_desired_in_minutes = steps_desired * minutes_per_timestep

        # initialize number of residence and heating by residents
        heating_by_residents = [0.0] * steps_desired_in_minutes
        number_of_residents = [0] * steps_desired_in_minutes

        # compute heat gains and number of persons
        for mode, gain in enumerate(gain_per_person):
            for timestep in range(steps_desired_in_minutes):
                number_of_residents[timestep] += occupancy_profile[mode]["Values"][timestep]
                heating_by_residents[timestep] = (
                    heating_by_residents[timestep] + gain * occupancy_profile[mode]["Values"][timestep]
                )
        if data_acquisition_mode == LpgDataAcquisitionMode.USE_UTSP:
            # load electricity consumption, water consumption and inner device heat gains
            electricity_data = io.StringIO(electricity)
            pre_electricity_consumption = pd.read_csv(
                electricity_data,
                sep=";",
                decimal=".",
                encoding="cp1252",
            ).loc[: (steps_desired_in_minutes - 1)]
            electricity_consumption_list = pd.to_numeric(
                pre_electricity_consumption["Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60W / min

            water_data = io.StringIO(warm_water)
            pre_water_consumption = pd.read_csv(
                water_data,
                sep=";",
                decimal=".",
                encoding="cp1252",
            ).loc[: (steps_desired_in_minutes - 1)]
            water_consumption_list = pd.to_numeric(pre_water_consumption["Sum [L]"]).tolist()

            inner_device_heat_gain_data = io.StringIO(inner_device_heat_gains)
            pre_inner_device_heat_gains = pd.read_csv(
                inner_device_heat_gain_data,
                sep=";",
                decimal=".",
                encoding="cp1252",
            ).loc[: (steps_desired_in_minutes - 1)]
            inner_device_heat_gains_list = pd.to_numeric(
                pre_inner_device_heat_gains["Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60W / min

        elif data_acquisition_mode == LpgDataAcquisitionMode.USE_PREDEFINED_PROFILE:
            # load electricity consumption, water consumption and inner device heat gains
            pre_electricity_consumption = pd.read_csv(
                electricity,
                sep=";",
                decimal=".",
                encoding="utf-8",
                usecols=["Sum [kWh]"],
            ).loc[: (steps_desired_in_minutes - 1)]
            electricity_consumption_list = pd.to_numeric(
                pre_electricity_consumption.loc[:, "Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60 000 W / min

            pre_water_consumption = pd.read_csv(
                warm_water,
                sep=";",
                decimal=".",
                encoding="utf-8",
                usecols=["Sum [L]"],
            ).loc[: (steps_desired_in_minutes - 1)]
            water_consumption_list = pd.to_numeric(pre_water_consumption.loc[:, "Sum [L]"]).tolist()

            pre_inner_device_heat_gains = pd.read_csv(
                inner_device_heat_gains,
                sep=";",
                decimal=".",
                encoding="utf-8",
                usecols=["Time", "Sum [kWh]"],
            ).loc[: (steps_desired_in_minutes - 1)]
            inner_device_heat_gains_list = pd.to_numeric(
                pre_inner_device_heat_gains.loc[:, "Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60W / min

        # put everything in a data frame and convert to utc
        initial_data = pd.DataFrame(
            {
                "Time": pd.date_range(
                    start=datetime.datetime(year=self.my_simulation_parameters.year, month=1, day=1),
                    end=datetime.datetime(year=self.my_simulation_parameters.year, month=1, day=1)
                    + datetime.timedelta(days=simulation_time_span.days)
                    - datetime.timedelta(seconds=60),
                    freq="min",
                )
            }
        )

        initial_data["number_of_residents"] = number_of_residents
        initial_data["heating_by_residents"] = heating_by_residents
        initial_data["electricity_consumption"] = electricity_consumption_list
        initial_data["water_consumption"] = water_consumption_list
        initial_data["inner_device_heat_gains"] = inner_device_heat_gains_list

        initial_data = utils.convert_lpg_data_to_utc(data=initial_data, year=self.my_simulation_parameters.year)

        # extract everything from data frame
        electricity_consumption = initial_data["electricity_consumption"].tolist()
        heating_by_residents = initial_data["heating_by_residents"].tolist()
        number_of_residents = initial_data["number_of_residents"].tolist()
        water_consumption = initial_data["water_consumption"].tolist()
        heating_by_devices = initial_data["inner_device_heat_gains"].tolist()

        # average data, when time resolution of inputs is coarser than time resolution of simulation
        if minutes_per_timestep > 1:
            # power needs averaging, not sum
            electricity_consumption = [
                sum(electricity_consumption[n : n + minutes_per_timestep]) / minutes_per_timestep
                for n in range(0, steps_desired_in_minutes, minutes_per_timestep)
            ]
            heating_by_devices = [
                sum(heating_by_devices[n : n + minutes_per_timestep]) / minutes_per_timestep
                for n in range(0, steps_desired_in_minutes, minutes_per_timestep)
            ]
            water_consumption = [
                sum(water_consumption[n : n + minutes_per_timestep])
                for n in range(0, steps_desired_in_minutes, minutes_per_timestep)
            ]
            heating_by_residents = [
                sum(heating_by_residents[n : n + minutes_per_timestep]) / minutes_per_timestep
                for n in range(0, steps_desired_in_minutes, minutes_per_timestep)
            ]
            number_of_residents = [
                int(sum(number_of_residents[n : n + minutes_per_timestep]) / minutes_per_timestep)
                for n in range(0, steps_desired_in_minutes, minutes_per_timestep)
            ]

        return (
            electricity_consumption,
            heating_by_devices,
            water_consumption,
            heating_by_residents,
            number_of_residents,
        )

    def cache_results(
        self,
        cache_filepath: str,
        number_of_residents: List,
        heating_by_residents: List,
        electricity_consumption: List,
        water_consumption: List,
        heating_by_devices: List,
        flexibility: Any,
        car_states: Any,
        car_locations: Any,
        driving_distances: Any,
        # saved_files: List,
    ) -> None:
        """Make caching file for the results."""

        # Saves normal loadprofile data in cache
        loadprofile_dataframe = pd.DataFrame(
            {
                "number_of_residents": number_of_residents,
                "heating_by_residents": heating_by_residents,
                "electricity_consumption": electricity_consumption,
                "water_consumption": water_consumption,
                "heating_by_devices": heating_by_devices,
            }
        )
        # save car data in cache
        car_data_dict = {
            "car_states": car_states,
            "car_locations": car_locations,
            "driving_distances": driving_distances,
        }
        # save flexibility data in cache
        flexibility_data_dict = {"flexibility": flexibility}

        # transform to dataframes
        car_dataframe = pd.DataFrame(car_data_dict)
        flexibility_dataframe = pd.DataFrame(flexibility_data_dict)

        # dump the dataframe to str i the cache content dict
        cache_content: Dict[str, Any] = {
            "data": loadprofile_dataframe,
            "car_data": car_dataframe,
            "flexibility_data": flexibility_dataframe,
        }

        for key, d_f in cache_content.items():
            cache_file = io.StringIO()
            d_f.to_csv(cache_file)
            d_f_str = cache_file.getvalue()
            cache_content.update({key: d_f_str})

        with open(cache_filepath, "w", encoding="utf-8") as file:
            json.dump(cache_content, file)
        del loadprofile_dataframe
        del car_dataframe
        del flexibility_dataframe

        log.information(f"Caching of lpg utsp results finished. Cache filepath is {cache_filepath}.")

    def load_results_and_transform_string_to_data(self, list_of_result_files: List[str]) -> List[Any]:
        """Transform a string of data into a data."""
        list_of_data = []
        for result_file_str in list_of_result_files:
            json_acceptable_string = result_file_str.replace("'", '"')
            if json_acceptable_string != "":
                data = json.loads(json_acceptable_string)
            else:
                data = {}
            list_of_data.append(data)
        return list_of_data

    def transform_dict_values(self, dict_to_check: Dict) -> Dict:
        """Function to convert string representations of data in a dictionary back to their original types."""
        transformed_data = {}
        for key, value in dict_to_check.items():
            try:
                # Attempt to evaluate the string value to a Python data type
                transformed_value = literal_eval(value)
            except (ValueError, SyntaxError):
                # If evaluation fails, keep the original string value
                transformed_value = value
            transformed_data[key] = transformed_value
        return transformed_data

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        occupancy_total_electricity_consumption_in_kilowatt_hour: Optional[float] = None
        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if output.field_name == self.ElectricityOutput:
                    occupancy_total_electricity_consumption_in_watt_series = postprocessing_results.iloc[:, index]
                    occupancy_total_electricity_consumption_in_kilowatt_hour = (
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=occupancy_total_electricity_consumption_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    break

        # make kpi entry
        occupancy_total_electricity_consumption_entry = KpiEntry(
            name="Residents' total electricity consumption",
            unit="kWh",
            value=occupancy_total_electricity_consumption_in_kilowatt_hour,
            tag=KpiTagEnumClass.RESIDENTS,
            description=self.component_name
        )

        list_of_kpi_entries = [occupancy_total_electricity_consumption_entry]
        return list_of_kpi_entries

    @staticmethod
    def get_cost_capex(config: UtspLpgConnectorConfig, simulation_parameters: SimulationParameters) -> cp.CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = cp.CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, snd write total energy consumption to component-config.

        No electricity costs for components except for Electricity Meter,
        because part of electricity consumption is feed by PV
        """
        consumption_in_kwh: float = 0.0
        for index, output in enumerate(all_outputs):
            if output.component_name == self.config.name and output.load_type == lt.LoadTypes.ELECTRICITY and output.field_name == self.ElectricityOutput:
                occupancy_total_electricity_consumption_in_watt_series = postprocessing_results.iloc[:, index]
                consumption_in_kwh = (
                KpiHelperClass.compute_total_energy_from_power_timeseries(
                    power_timeseries_in_watt=occupancy_total_electricity_consumption_in_watt_series,
                    timeresolution=self.my_simulation_parameters.seconds_per_timestep)
                )

        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=0,
            co2_footprint_in_kg=0,
            consumption_in_kwh=consumption_in_kwh,
            loadtype=lt.LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.RESIDENTS
        )

        return opex_cost_data_class



# clean

""" Computes all relevant parameters for data base of Energy System Models, as well as parameters needed for validation of the building types."""

import datetime as dt
import os
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from hisim import log, utils
from hisim.component_wrapper import ComponentWrapper
from hisim.loadtypes import ComponentType, HeatingSystems, InandOutputType, LoadTypes
from hisim.simulationparameters import SimulationParameters
from hisim.components.loadprofilegenerator_utsp_connector import UtspLpgConnectorConfig
from hisim.components.weather import Weather


def compute_energy_from_power(power_timeseries: pd.Series, seconds_per_timestep: int) -> float:
    """Computes energy from power value."""
    if power_timeseries.empty:
        return 0.0
    return float(power_timeseries.sum() * seconds_per_timestep / 3.6e6)


# TODO: enable this functio for the lpg_utsp_connector if needed
# so far he lpg_utsp_connector only has the predefined profile CHR01 and not the AVG profiles
# def get_factor_cooking(occupancy_config: UtspLpgConnectorConfig) -> float:
#     """Reads in portion of electricity consumption which is assigned to cooking."""

#     if occupancy_config.profile_name != "AVG":
#         return 0
#     scaling_factors = pd.read_csv(
#         utils.HISIMPATH["occupancy_scaling_factors_per_country"], encoding="utf-8", sep=";", index_col=1,
#     )
#     if occupancy_config.country_name in scaling_factors.index:
#         scaling_factor_line = scaling_factors.loc[occupancy_config.country_name]
#     else:
#         scaling_factor_line = scaling_factors.loc["EU"]
#         log.warning(
#             "Scaling Factor for " + occupancy_config.country_name + "is not available, EU average is used per default."
#         )
#     return float(scaling_factor_line["ratio cooking to total"])


def compute_seasonal(
    csv_frame_seasonal: pd.DataFrame,
    simulation_parameters: SimulationParameters,
    index_in_seasonal_frame: Tuple[str, str],
    factor: float,
    output: pd.Series,
    day: pd.Series,
    night: pd.Series,
) -> pd.DataFrame:
    """Takes annual time series and computes average consumption during day/night in summer, winter and intermediate time.

    Summer from daylight 21.06.2019 to daylight 23.09.2019 (total 94 days)
    Winter from 01.01.2019 - daylight 23.03.2019 and daylight 21.12.2019 to 31.12.2019 (total 92 days)
    Intermediate from daylight 23.03.2029 - daylight 21.06.2019 and daylight 23.09.2019 - daylight 21.12.2019 (total 179)
    """

    output_day = output[day.index]
    output_night = output[night.index]
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Summer-Day"] = (
        output_day[
            (
                (output_day.index > dt.datetime(year=simulation_parameters.year, month=6, day=21, hour=0))
                & (output_day.index < dt.datetime(year=simulation_parameters.year, month=9, day=23, hour=0))
            )
        ].sum()
        * factor
    )
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Summer-Night"] = (
        output_night[
            (
                (output_night.index > dt.datetime(year=simulation_parameters.year, month=6, day=21, hour=12))
                & (output_night.index < dt.datetime(year=simulation_parameters.year, month=9, day=23, hour=12))
            )
        ].sum()
        * factor
    )
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Winter-Day"] = (
        output_day[
            (
                (output_day.index > dt.datetime(year=simulation_parameters.year, month=12, day=21, hour=0))
                | (output_day.index < dt.datetime(year=simulation_parameters.year, month=3, day=23, hour=0))
            )
        ].sum()
        * factor
    )
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Winter-Night"] = (
        output_night[
            (
                (output_night.index > dt.datetime(year=simulation_parameters.year, month=12, day=21, hour=12))
                | (output_night.index < dt.datetime(year=simulation_parameters.year, month=3, day=23, hour=12))
            )
        ].sum()
        * factor
    )
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Intermediate-Day"] = (
        output_day[
            (
                (output_day.index > dt.datetime(year=simulation_parameters.year, month=3, day=23, hour=0))
                & (output_day.index < dt.datetime(year=simulation_parameters.year, month=6, day=21, hour=0))
            )
            | (
                (output_day.index > dt.datetime(year=simulation_parameters.year, month=9, day=23, hour=0))
                & (output_day.index < dt.datetime(year=simulation_parameters.year, month=12, day=21, hour=0))
            )
        ].sum()
        * factor
    )
    csv_frame_seasonal.loc[index_in_seasonal_frame, "Intermediate-Night"] = (
        output_night[
            (
                (output_night.index > dt.datetime(year=simulation_parameters.year, month=3, day=23, hour=12))
                & (output_night.index < dt.datetime(year=simulation_parameters.year, month=6, day=21, hour=12))
            )
            | (
                (output_night.index > dt.datetime(year=simulation_parameters.year, month=9, day=23, hour=12))
                & (output_night.index < dt.datetime(year=simulation_parameters.year, month=12, day=21, hour=12))
            )
        ].sum()
        * factor
    )

    return csv_frame_seasonal


def generate_csv_for_database(
    all_outputs: List,
    results: pd.DataFrame,
    simulation_parameters: SimulationParameters,
    building_data: pd.DataFrame,
    occupancy_config: Optional[UtspLpgConnectorConfig],
    wrapped_components: List[ComponentWrapper],
) -> None:
    """Extracts relevant data from the HiSIM simulation and puts it together in a .csv file.

    :param all_outputs: List of all outputs
    :type all_outputs: List
    :param results: DataFrame containing all outputs at all time steps
    :type results: pd.DataFrame
    :param simulation_parameters: parameters of the simulation
    :type simulation_parameters: SimulationParameters
    :param building_data: tabula data of building type
    :type building_data: pd.DataFrame
    """

    fuels_heating = [
        "Oil [l]",
        "Gas [kWh]",
        "Distributed Stream [kWh]",
        "Solar [kWh]",
        "Electricity [kWh]",
        "Electricity - HeatPump [kWh]",
        "FuelCell [kWh]",
    ]
    fuels_air_cooling = ["Electricity [kWh]"]
    fuels_cooking = ["Gas [kWh]", "Electricity [kWh]"]
    fuels_mobility = ["Diesel [l]", "Electricity [kWh]"]
    fuels_remaining_load = ["Electricity [kWh]"]
    fuels_annual_heating_demand = ["[kWh/(m*m*a)]"]
    fuels_building_size = ["Area [m*m]"]
    fuels_construction_year = ["Period start", "Period end"]
    fuels_heating_days = ["Number of Days"]
    fuels_average_temperature = ["Temperature [C]"]

    device_index = (
        ["SpaceHeating"] * len(fuels_heating)
        + ["WaterHeating"] * len(fuels_heating)
        + ["AirCooling"] * len(fuels_air_cooling)
        + ["Cooking"] * len(fuels_cooking)
        + ["Transport"] * len(fuels_mobility)
        + ["RemainingLoad"] * len(fuels_remaining_load)
        + ["Annual Heating Demand Tabula"]
        + ["Annual Heating Demand HiSIM"]
        + ["Annual Heating Demand Tabula with HiSIM climate"]
        + ["Building Size"]
        + ["Construction Year"] * 2
        + ["HeatingDays Tabula"]
        + ["HeatingDays HiSIM"]
        + ["AverageTemperatureInHeatingSeason Tabula"]
        + ["AverageTemperatureInHeatingSeason HiSIM"]
    )
    units = (
        fuels_heating * 2
        + fuels_air_cooling
        + fuels_cooking
        + fuels_mobility
        + fuels_remaining_load
        + fuels_annual_heating_demand * 3
        + fuels_building_size
        + fuels_construction_year
        + fuels_heating_days * 2
        + fuels_average_temperature * 2
    )
    tuples = list(zip(*[device_index, units]))

    csv_frame_annual = pd.Series(
        [0.0] * len(device_index),
        index=pd.MultiIndex.from_tuples(tuples, names=["Category", "Fuel"]),
    )
    csv_frame_seasonal = pd.DataFrame(
        {
            "Summer-Day": [0.0] * (len(device_index) - 10),
            "Summer-Night": [0.0] * (len(device_index) - 10),
            "Winter-Day": [0.0] * (len(device_index) - 10),
            "Winter-Night": [0.0] * (len(device_index) - 10),
            "Intermediate-Day": [0.0] * (len(device_index) - 10),
            "Intermediate-Night": [0.0] * (len(device_index) - 10),
        },
        index=pd.MultiIndex.from_tuples(tuples[:-10], names=["Category", "Fuel"]),
    )

    remaining_electricity_annual = 0.0
    remaining_electricity_seasonal = np.array([0.0] * 6)
    day = None
    night = None
    # get indices for day and night:
    for index, output in enumerate(all_outputs):
        for elem in wrapped_components:
            if isinstance(elem.my_component, Weather):
                if output.component_name == elem.my_component.component_name and output.field_name == "Altitude":
                    altitude_data = results.iloc[:, index]
                    day = altitude_data[altitude_data > 0]
                    night = altitude_data[altitude_data <= 0]

    buildings_in_district_list = building_data.index.to_list()
    for building_object in buildings_in_district_list:
        for index, output in enumerate(all_outputs):
            if building_object in str(output.get_pretty_name()) and output.postprocessing_flag is not None:
                if InandOutputType.WATER_HEATING in output.postprocessing_flag:
                    if LoadTypes.DISTRICTHEATING in output.postprocessing_flag:
                        csv_frame_annual[("WaterHeating", "Distributed Stream [kWh]")] = (
                            sum(results.iloc[:, index]) * 1e-3
                        )
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=(
                                "WaterHeating",
                                "Distributed Stream [kWh]",
                            ),
                            factor=1e-3,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                    elif LoadTypes.GAS in output.postprocessing_flag:
                        csv_frame_annual[("WaterHeating", "Gas [kWh]")] = sum(results.iloc[:, index]) * 1e-3
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("WaterHeating", "Gas [kWh]"),
                            output=results.iloc[:, index],
                            factor=1e-3,
                            day=day,
                            night=night,
                        )
                    elif LoadTypes.OIL in output.postprocessing_flag:
                        csv_frame_annual[("WaterHeating", "Oil [l]")] = sum(results.iloc[:, index])
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("WaterHeating", "Oil [l]"),
                            output=results.iloc[:, index],
                            factor=1,
                            day=day,
                            night=night,
                        )
                    else:
                        if HeatingSystems.HEAT_PUMP in output.postprocessing_flag:
                            csv_frame_annual[("WaterHeating", "Electricity - HeatPump [kWh]")] = (
                                compute_energy_from_power(
                                    power_timeseries=results.iloc[:, index],
                                    seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                                )
                            )
                            csv_frame_seasonal = compute_seasonal(
                                simulation_parameters=simulation_parameters,
                                csv_frame_seasonal=csv_frame_seasonal,
                                index_in_seasonal_frame=(
                                    "WaterHeating",
                                    "Electricity - HeatPump [kWh]",
                                ),
                                factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                                output=results.iloc[:, index],
                                day=day,
                                night=night,
                            )
                        elif HeatingSystems.ELECTRIC_HEATING in output.postprocessing_flag:
                            csv_frame_annual[("WaterHeating", "Electricity [kWh]")] = compute_energy_from_power(
                                power_timeseries=results.iloc[:, index],
                                seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                            )
                            csv_frame_seasonal = compute_seasonal(
                                simulation_parameters=simulation_parameters,
                                csv_frame_seasonal=csv_frame_seasonal,
                                index_in_seasonal_frame=(
                                    "WaterHeating",
                                    "Electricity [kWh]",
                                ),
                                factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                                output=results.iloc[:, index],
                                day=day,
                                night=night,
                            )
                elif InandOutputType.HEATING in output.postprocessing_flag:
                    if LoadTypes.DISTRICTHEATING in output.postprocessing_flag:
                        csv_frame_annual[("SpaceHeating", "Distributed Stream [kWh]")] = (
                            sum(results.iloc[:, index]) * 1e-3
                        )
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=(
                                "SpaceHeating",
                                "Distributed Stream [kWh]",
                            ),
                            factor=1e-3,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                    elif LoadTypes.GAS in output.postprocessing_flag:
                        csv_frame_annual[("SpaceHeating", "Gas [kWh]")] = sum(results.iloc[:, index]) * 1e-3
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("SpaceHeating", "Gas [kWh]"),
                            factor=1e-3,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                    elif LoadTypes.OIL in output.postprocessing_flag:
                        csv_frame_annual[("SpaceHeating", "Oil [l]")] = sum(results.iloc[:, index])
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("SpaceHeating", "Oil [l]"),
                            factor=1,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                    else:
                        if HeatingSystems.HEAT_PUMP in output.postprocessing_flag:
                            csv_frame_annual[("SpaceHeating", "Electricity - HeatPump [kWh]")] = (
                                compute_energy_from_power(
                                    power_timeseries=results.iloc[:, index],
                                    seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                                )
                            )
                            csv_frame_seasonal = compute_seasonal(
                                simulation_parameters=simulation_parameters,
                                csv_frame_seasonal=csv_frame_seasonal,
                                index_in_seasonal_frame=(
                                    "SpaceHeating",
                                    "Electricity - HeatPump [kWh]",
                                ),
                                factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                                output=results.iloc[:, index],
                                day=day,
                                night=night,
                            )
                        elif HeatingSystems.ELECTRIC_HEATING in output.postprocessing_flag:
                            csv_frame_annual[("SpaceHeating", "Electricity [kWh]")] = compute_energy_from_power(
                                power_timeseries=results.iloc[:, index],
                                seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                            )
                            csv_frame_seasonal = compute_seasonal(
                                simulation_parameters=simulation_parameters,
                                csv_frame_seasonal=csv_frame_seasonal,
                                index_in_seasonal_frame=(
                                    "SpaceHeating",
                                    "Electricity [kWh]",
                                ),
                                factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                                output=results.iloc[:, index],
                                day=day,
                                night=night,
                            )
                elif ComponentType.CAR in output.postprocessing_flag:
                    if LoadTypes.DIESEL in output.postprocessing_flag:
                        csv_frame_annual[("Transport", "Diesel [l]")] = sum(results.iloc[:, index])
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("Transport", "Diesel [l]"),
                            factor=1,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                    else:
                        csv_frame_annual[("Transport", "Electricity [kWh]")] = compute_energy_from_power(
                            power_timeseries=results.iloc[:, index],
                            seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                        )
                        csv_frame_seasonal = compute_seasonal(
                            simulation_parameters=simulation_parameters,
                            csv_frame_seasonal=csv_frame_seasonal,
                            index_in_seasonal_frame=("Transport", "Electricity [kWh]"),
                            factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                            output=results.iloc[:, index],
                            day=day,
                            night=night,
                        )
                elif InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED in output.postprocessing_flag:
                    remaining_electricity_annual = remaining_electricity_annual + compute_energy_from_power(
                        power_timeseries=results.iloc[:, index],
                        seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                    )
                    remaining_electricity_seasonal_item = compute_seasonal(
                        simulation_parameters=simulation_parameters,
                        csv_frame_seasonal=csv_frame_seasonal,
                        index_in_seasonal_frame=("RemainingLoad", "Electricity [kWh]"),
                        factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                        output=results.iloc[:, index],
                        day=day,
                        night=night,
                    ).loc[("RemainingLoad", "Electricity [kWh]")]
                    remaining_electricity_seasonal = remaining_electricity_seasonal + np.array(
                        remaining_electricity_seasonal_item
                    )
                elif InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED in output.postprocessing_flag:
                    remaining_electricity_annual = remaining_electricity_annual + compute_energy_from_power(
                        power_timeseries=results.iloc[:, index],
                        seconds_per_timestep=simulation_parameters.seconds_per_timestep,
                    )
                    remaining_electricity_seasonal_item = compute_seasonal(
                        simulation_parameters=simulation_parameters,
                        csv_frame_seasonal=csv_frame_seasonal,
                        index_in_seasonal_frame=("RemainingLoad", "Electricity [kWh]"),
                        factor=simulation_parameters.seconds_per_timestep / 3.6e6,
                        output=results.iloc[:, index],
                        day=day,
                        night=night,
                    ).loc[("RemainingLoad", "Electricity [kWh]")]
                    remaining_electricity_seasonal = remaining_electricity_seasonal + np.array(
                        remaining_electricity_seasonal_item
                    )
                else:
                    continue

        if occupancy_config is None:
            factor_cooking = 0.0
        else:
            log.warning(
                "The factor_cooking can not be calcuated with the lpg_utsp_connector. "
                + "The AVG profiles must be implemented in the lpg_ustp_connector first."
            )
            factor_cooking = 0.0  # get_factor_cooking(occupancy_config)

        csv_frame_annual[("RemainingLoad", "Electricity [kWh]")] = remaining_electricity_annual * (1 - factor_cooking)
        csv_frame_annual[("Cooking", "Electricity [kWh]")] = remaining_electricity_annual * factor_cooking
        csv_frame_seasonal.loc[("RemainingLoad", "Electricity [kWh]")] = remaining_electricity_seasonal * (
            1 - factor_cooking
        )
        csv_frame_seasonal.loc[("Cooking", "Electricity [kWh]")] = remaining_electricity_seasonal * factor_cooking

        # extract infos from used climate data to compare to climate information used for tabula evaluation
        building_code = building_data.loc[building_object, "Code_BuildingVariant"]
        converting_data = pd.read_csv(utils.HISIMPATH["housing_reference_temperatures"])
        converting_data.index = converting_data["Location"]  # type: ignore

        # write all necesary data for building validation to csv file
        csv_frame_annual[("Annual Heating Demand Tabula", "[kWh/(m*m*a)]")] = building_data.loc[building_object, "q_ht"]
        csv_frame_annual[("HeatingDays Tabula", "Number of Days")] = building_data.loc[building_object, "HeatingDays"]
        csv_frame_annual[("AverageTemperatureInHeatingSeason Tabula", "Temperature [C]")] = building_data.loc[
            building_object, "Theta_e"
        ]
        csv_frame_annual[("Annual Heating Demand HiSIM", "[kWh/(m*m*a)]")] = (
            csv_frame_annual[("SpaceHeating", "Distributed Stream [kWh]")]
            / building_data.loc[building_object, "A_C_Ref"]
        )
        csv_frame_annual[("HeatingDays HiSIM", "Number of Days")] = int(
            converting_data.loc[building_code.split(".")[0]]["NumberOfHeatingDays"]
        )
        csv_frame_annual[("AverageTemperatureInHeatingSeason HiSIM", "Temperature [C]")] = float(
            converting_data.loc[building_code.split(".")[0]]["Average"]
        )
        csv_frame_annual[("Building Size", "Area [m*m]")] = building_data.loc[building_object, "A_C_Ref"]
        csv_frame_annual[("Construction Year", "Period start")] = building_data.loc[building_object, "Year1_Building"]
        csv_frame_annual[("Construction Year", "Period end")] = building_data.loc[building_object, "Year2_Building"]
        csv_frame_annual[("Annual Heating Demand Tabula with HiSIM climate", "[kWh/(m*m*a)]")] = (
            building_data.loc[building_object, "q_ht"]
            * (
                (20 - csv_frame_annual[("AverageTemperatureInHeatingSeason HiSIM", "Temperature [C]")])
                * csv_frame_annual[("HeatingDays HiSIM", "Number of Days")]
            )
            / (
                (20 - csv_frame_annual[("AverageTemperatureInHeatingSeason Tabula", "Temperature [C]")])
                * csv_frame_annual[("HeatingDays Tabula", "Number of Days")]
            )
        )

        csv_frame_annual.to_csv(
            os.path.join(
                simulation_parameters.result_directory,
                f"csv_for_housing_data_base_annual_{building_object}.csv",
            ),
            encoding="utf-8",
        )
        csv_frame_seasonal.to_csv(
            os.path.join(
                simulation_parameters.result_directory,
                f"csv_for_housing_data_base_seasonal_{building_object}.csv",
            ),
            encoding="utf-8",
        )



"""Webtool results with important kpis."""

from dataclasses import dataclass, field, InitVar
from typing import Dict, List, Optional

import pandas as pd
from dataclass_wizard import JSONWizard

from hisim.component import ComponentOutput
from hisim.loadtypes import OutputPostprocessingRules
from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer


@dataclass
class ResultEntry(JSONWizard):
    """Class for storing one result entry for hisim webtool."""

    name: str
    unit: str
    value: float
    description: Optional[str] = None


@dataclass
class WebtoolDict(JSONWizard):
    """Class for storing results for hisim webtool."""

    kpis: Dict
    components: Dict = field(init=False)

    post_processing_data_transfer: InitVar[PostProcessingDataTransfer]
    computed_opex: InitVar[List]
    computed_capex: InitVar[List]

    def __post_init__(self, post_processing_data_transfer, computed_opex, computed_capex):
        """Build the dataclass from input data."""
        self.components = {}
        self.init_structure(post_processing_data_transfer)
        self.add_opex_capex_results(computed_opex, computed_capex)
        self.add_technical_results(post_processing_data_transfer)
        self.add_configuration(post_processing_data_transfer)

    def init_structure(self, post_processing_data_transfer):
        """Initialize results dict for webtool with component names and categories.

        Only components with DisplayConfig.display_in_webtool=True are selected.
        """

        # Get bools that tells if the components should be displayed in webtool
        for component in post_processing_data_transfer.wrapped_components:
            if component.my_component.my_display_config.display_in_webtool:
                this_name = component.my_component.component_name
                this_pretty_name = component.my_component.my_display_config.pretty_name
                this_component_class = component.my_component.get_classname()
                self.components[this_name] = {
                    "prettyName": this_pretty_name,
                    "class": this_component_class,
                    "configuration": {},
                    "operation": {},
                    "economics": {},
                }

    def add_opex_capex_results(self, computed_opex, computed_capex):
        """Add results from the results of `opex_calculation()` and `capex_calculation()` to webtool dict."""
        categories_opex = ["economics", "operation", "economics"]
        categories_capex = ["economics", "operation", "economics"]
        # Get OPEX and CAPEX
        for computed_values, categories in zip([computed_opex, computed_capex], [categories_opex, categories_capex]):
            if not all(isinstance(_, str) for _ in computed_values[0]):
                # First row is header.
                raise ValueError("Expected header in first row.")
            for computed_values_row in computed_values[1:]:
                if not isinstance(computed_values_row[0], str):
                    # Fist column is component name.
                    raise ValueError("Expected component name in first column.")
                if "total" in computed_values_row[0].lower():
                    # Skip rows with total values.
                    continue
                for idx_column, computed_values_item in enumerate(computed_values_row[1:]):
                    # Get component name
                    this_component = computed_values_row[0]
                    # Skip components that should not be displayed
                    if this_component not in self.components.keys():
                        continue
                    # Get value key and reformat unit into brackets
                    computed_values_key = computed_values[0][idx_column + 1]
                    computed_values_name, computed_values_unit = computed_values_key.split(" in ")
                    # Create result entry
                    result_entry = ResultEntry(
                        name=computed_values_name, unit=computed_values_unit, description="", value=computed_values_item
                    )
                    # Save to dict
                    self.components[this_component][categories[idx_column]].update({computed_values_name: result_entry})

    def add_technical_results(self, post_processing_data_transfer: PostProcessingDataTransfer) -> None:
        """Add technical results from PostProcessingDataTransfer to results dataclass."""

        # Get outputs
        name_component_dict: Dict[str, ComponentOutput] = {}
        component_output: ComponentOutput
        for component_output in post_processing_data_transfer.all_outputs:
            name_component_dict[component_output.get_pretty_name()] = component_output

        # Get bools that tells if the components should be displayed in webtool
        component_display_in_webtool_dict: Dict[str, bool] = {}
        for component in post_processing_data_transfer.wrapped_components:
            component_display_in_webtool_dict[
                component.my_component.component_name
            ] = component.my_component.my_display_config.display_in_webtool

        # Read data from PostProcessingDataTransfer and save to results
        data: pd.DataFrame = post_processing_data_transfer.results_cumulative
        series_name: str
        series: pd.Series
        for series_name, series in data.items():
            my_output = name_component_dict[series_name]
            # Skip components that should not be displayed
            if my_output.component_name not in self.components.keys():
                continue
            # Skip outputs without postprocessing flag OutputPostprocessingRules.DISPLAY_IN_WEBTOOL
            if my_output.postprocessing_flag:
                if OutputPostprocessingRules.DISPLAY_IN_WEBTOOL not in my_output.postprocessing_flag:
                    continue
            else:
                continue
            # Create result entry
            my_result = ResultEntry(
                name=my_output.display_name,
                unit=my_output.unit,
                description=my_output.output_description,
                value=round(series.item(), 2),
            )
            # Write to result dict
            self.components[my_output.component_name]["operation"].update({my_output.display_name: my_result.to_dict()})

    def add_configuration(self, ppdt: PostProcessingDataTransfer) -> None:
        """Add configuration for displayed components from PostProcessingDataTransfer to results dataclass."""
        for wrapped_component in ppdt.wrapped_components:
            if wrapped_component.my_component.my_display_config.display_in_webtool:
                component_content = wrapped_component.my_component.config.to_dict()
                self.components[wrapped_component.my_component.component_name]["configuration"] = component_content



"""Generic ev charger module."""

# clean

# Generic/Built-in
import os
from typing import Any, List, Optional
import json
import copy
import sqlite3
import datetime
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import pandas as pd

# Owned
from hisim.simulationparameters import SimulationParameters
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils


__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class VehiclePureConfig(cp.ConfigBase):
    """Vehicle Pure Config class."""

    building_name: str
    name: str
    manufacturer: str
    model: str
    soc: float
    profile: str

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return VehiclePure.get_full_classname()

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return VehiclePureConfig(
            building_name=building_name,
            name="Electrical Charger",
            manufacturer="Tesla",
            model="Model 3 v3",
            soc=1.0,
            profile="CHR01",
        )


@dataclass_json
@dataclass
class EVChargerControllerConfig(cp.ConfigBase):
    """Electrical vehicle charger controller config class."""

    building_name: str
    name: str
    mode: int

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return EVChargerController.get_full_classname()

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return EVChargerControllerConfig(
            name="ElectricalChargerController",
            mode=1,
            building_name=building_name,
        )


@dataclass_json
@dataclass
class VehicleConfig(cp.ConfigBase):
    """Vehicle config class."""

    building_name: str
    name: str
    manufacturer: str
    model: str
    soc: float

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return Vehicle.get_full_classname()

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return VehicleConfig(
            building_name=building_name,
            name="ElectricVehicle",
            manufacturer="Renault",
            model="Zoe v3",
            soc=0.8,
        )


@dataclass_json
@dataclass
class EVChargerConfig(cp.ConfigBase):
    """Electrical vehicle config class."""

    building_name: str
    name: str
    manufacturer: str
    charger_name: str
    electric_vehicle: Any

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return EVCharger.get_full_classname()

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return EVChargerConfig(
            building_name=building_name,
            name="EV_Charger",
            manufacturer="myenergi",
            charger_name="Wallbox ZAPPI 222TW",
            electric_vehicle=None,
        )


class VehiclePure(cp.Component):
    """Vehicle component class.

    Parameters
    ----------
    manufacturer : str
        Electric vehicle Manufacturer
    model : str
        Electric vehicle model
    soc : float
        Initial state of charge of battery before
        simulation
    profile : str
        Family profile imported from LPG. The family
        denomination defines the electric vehicle usage
        throughout the simulation duration

    """

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: VehiclePureConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.evconfig = config

        self.build()

    def build(self) -> None:
        """Build function."""
        if self.evconfig.soc > 1 or self.evconfig.soc < 0:
            raise Exception("Invalid State Of Charge.")

        # Gets flexibilities, including heat pump
        electric_vehicle_database = utils.load_smart_appliance("Electric Vehicle")

        electric_vehicle_found = False
        electric_vehicle = None
        for electric_vehicle in electric_vehicle_database:
            if (
                electric_vehicle["Manufacturer"] == self.evconfig.manufacturer
                and electric_vehicle["Model"] == self.evconfig.model
            ):
                electric_vehicle_found = True
                break

        if not electric_vehicle_found or electric_vehicle is None:
            raise Exception("Electric Vehicle not registered in the database")

        self.max_capacity = electric_vehicle["Battery Capacity"] * 1e3
        if "Battery Useful Capacity" in electric_vehicle:
            self.min_capacity = self.max_capacity - electric_vehicle["Battery Useful Capacity"] * 1e3
        else:
            self.min_capacity = self.max_capacity * 0.1

        capacity = self.max_capacity * self.evconfig.soc
        if capacity < self.min_capacity:
            capacity = self.min_capacity

        self.capacity = capacity

        cache_file_exists, cache_filepath = utils.get_cache_file(
            self.component_name, self.evconfig, self.my_simulation_parameters
        )
        if cache_file_exists:
            self.car_in_charging_station = pd.read_csv(cache_filepath, sep=",", decimal=".")[
                "CarInChargingStation"
            ].tolist()
            self.discharge = pd.read_csv(cache_filepath, sep=",", decimal=".")["Discharge"].tolist()
        else:

            def open_sql(path, table_name):
                sql_file = sqlite3.connect(path)
                return pd.read_sql(f"SELECT * FROM {table_name};", sql_file)

            def open_ev_json(filepath):
                with open(filepath, encoding="utf-8") as file:
                    data = json.load(file)
                return data["Values"]

            filepath = utils.load_export_load_profile_generator(target=self.evconfig.profile_name)
            if filepath is None:
                filepath = utils.HISIMPATH

            ev_files = {}
            filepaths = open_sql(filepath["electric_vehicle"][1], "ResultFileEntries")
            list_columns = []
            list_values = []
            for _, row in filepaths.iterrows():
                json_info = json.loads(row["Json"])
                if "Charging" in json_info["FileName"] and "png" not in json_info["FileName"]:
                    filepath = os.path.normpath(json_info["FullFileName"])
                    filepath_list = filepath.split(os.sep)
                    ev_files[filepath_list[-1].split(".")[0]] = json_info["FullFileName"]
                    list_columns.append(filepath_list[-1].split(".")[0])
                    list_values.append(open_ev_json(json_info["FullFileName"]))
            list_values = list(map(list, zip(*list_values)))
            # ev_pd = pd.DataFrame(list_values, columns=list_columns)

            # Gets battery information to calculate discharging while not at home
            transportation_devices = open_sql(filepath["electric_vehicle"][0], "TransportationDevices")
            for _, vehicle in transportation_devices.iterrows():
                if "Charging" in vehicle["Name"]:
                    vehicle_info = json.loads(vehicle["Json"])
                    battery_stored_energy_meters = vehicle_info["FullRangeInMeters"]
                    convert_factor = vehicle_info["EnergyToDistanceFactor"]
                    # battery_stored_energy_wh = battery_stored_energy_meters * convert_factor

            convert_factor_meters_by_wh = convert_factor * 3600
            soc = []
            load_stats = []
            car_in_charging_station = []
            # car_state = []
            discharge_stats = [0]
            # Gets transportation stats
            transportation_devices_stats = open_sql(filepath["electric_vehicle"][0], "TransportationDeviceStates")
            for _, column in transportation_devices_stats.iterrows():
                if datetime.datetime.strptime(column["DateTime"], "%d/%m/%Y %H:%M") > datetime.datetime.strptime(
                    "31/12/2018 23:59", "%d/%m/%Y %H:%M"
                ):
                    if (
                        "ParkingAndFullyCharged" in column["DeviceState"]
                        or "ParkingAndCharging" in column["DeviceState"]
                    ):
                        car_in_charging_station.append(True)
                    else:
                        car_in_charging_station.append(False)
                    load_stats.append(float(column["CurrentRange"]))
                    soc.append(float(column["CurrentRange"]) / battery_stored_energy_meters)
                    if len(load_stats) > 1:
                        diff = load_stats[-1] - load_stats[-2]
                        if diff < 0:
                            discharge_stats.append(diff / convert_factor_meters_by_wh)
                        else:
                            discharge_stats.append(0)

            data: List = []
            data.append(car_in_charging_station)
            data.append(discharge_stats)
            data = list(map(list, zip(*data)))

            data_parameters = ["CarInChargingStation", "Discharge"]
            database = pd.DataFrame(data, columns=data_parameters)

            self.car_in_charging_station = car_in_charging_station
            self.discharge = discharge_stats
            database.to_csv(cache_filepath)
            # utils.save_cache("Vehicle", [self.evconfig.profile_name], database)

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        pass


class Vehicle(cp.Component):
    """Electric Vehicle Component.

    This is a alternative implementation, not fully working,
    to be used in future releases.

    Parameters
    ----------
    manufacturer : str
        Electric vehicle Manufacturer
    model : str
        Electric vehicle model
    soc : float
        Initial state of charge of battery before
        simulation

    """

    BeforeCapacity = "BeforeCapacity"

    AfterCapacity = "AfterCapacity"
    MaxCapacity = "MaxCapacity"
    Discharge = "Discharge"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: VehicleConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(manufacturer=config.manufacturer, model=config.model, soc=config.soc)

        self.before_capacity_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BeforeCapacity,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )

        self.after_capacity_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.AfterCapacity, lt.LoadTypes.ANY, lt.Units.WATT
        )
        self.max_capacity_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.MaxCapacity, lt.LoadTypes.ANY, lt.Units.WATT
        )
        self.discharge_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.Discharge, lt.LoadTypes.ELECTRICITY, lt.Units.WATT
        )

    def build(self, manufacturer: str, model: str, soc: float) -> None:
        """Build function."""
        if soc > 1 or soc < 0:
            raise Exception("Invalid State Of Charge.")

        # Gets flexibilities, including heat pump
        electric_vehicle_database = utils.load_smart_appliance("Electric Vehicle")

        electric_vehicle_found = False
        electric_vehicle = None
        for electric_vehicle in electric_vehicle_database:
            if electric_vehicle["Manufacturer"] == manufacturer and electric_vehicle["Model"] == model:
                electric_vehicle_found = True
                break

        if not electric_vehicle_found or electric_vehicle is None:
            raise Exception("Electric Vehicle not registered in the database")

        self.max_capacity = electric_vehicle["Battery Capacity"] * 1e3
        if "Battery Useful Capacity" in electric_vehicle:
            self.min_capacity = self.max_capacity - electric_vehicle["Battery Useful Capacity"] * 1e3
        else:
            self.min_capacity = self.max_capacity * 0.1

        capacity = self.max_capacity * soc
        if capacity < self.min_capacity:
            capacity = self.min_capacity

        self.capacity = capacity

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doubelchecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        if timestep == 0:
            capacity = self.capacity
        else:
            capacity = stsv.get_input_value(self.before_capacity_channel)
        stsv.set_output_value(self.after_capacity_channel, capacity)
        stsv.set_output_value(self.max_capacity_channel, self.max_capacity)


class SimpleStorageState:
    """Simple Storage State class.

    Simplistic implementation for any type
    of energy state storage. Relevant for battery,
    electric vehicles, etc, to store or withdraw
    certain amount of energy of a predefined storage.

    Parameters
    ----------
    max_var_val: float
        Maximum variation energy, i.e., maximum power
    min_var_val: float
        Minimum variation energy, i.e., minimum power
    stored_energy : float
        Initial storage energy
    time_correction_factor : float
        Time correction factor to convert power into energy
        the time step duration used in the simulation. This
        factor is equivalent of the invert of seconds per time
        step.
    seconds_per_timestep : int
        Duration in seconds of one time step.

    """

    def __init__(
        self,
        max_var_val: float,
        min_var_val: float,
        stored_energy: float = 0.0,
        time_correction_factor: Optional[float] = None,
        seconds_per_timestep: Optional[int] = None,
    ) -> None:
        """Initialize the class."""
        self.max_var_val = max_var_val
        self.min_var_val = min_var_val
        self.stored_energy = stored_energy
        self.time_correction_factor = time_correction_factor
        self.seconds_per_timestep = seconds_per_timestep

    def store(
        self,
        max_capacity: float,
        current_capacity: float,
        val: float,
        efficiency: float = 1.0,
    ) -> Any:
        """Store."""
        if val < max_capacity - current_capacity:
            amount = val
        else:
            amount = max_capacity - current_capacity

        # Check if the theoretical capacity is larger than maximum capacity
        if current_capacity >= max_capacity:
            amount = 0

        # Check if the charging variation is larger than physical possible
        if amount > self.max_var_val:
            amount = self.max_var_val

        amount = amount * efficiency

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def force_store(self, max_capacity: float, current_capacity: float) -> Any:
        """Force store."""
        amount = self.max_var_val
        if amount > max_capacity - current_capacity:
            amount = max_capacity - current_capacity

        # Check if the theoretical capacity is larger than maximum capacity
        if current_capacity >= max_capacity:
            amount = 0

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def withdraw(
        self,
        min_capacity: float,
        current_capacity: float,
        val: float,
        efficiency: float = 1.0,
    ) -> Any:
        """Withdraw."""
        val = abs(val)
        if current_capacity - min_capacity > val:
            amount = -val
        else:
            amount = min_capacity - current_capacity

        if current_capacity <= min_capacity:
            amount = 0

        if amount < self.min_var_val:
            amount = self.min_var_val

        amount = amount / efficiency

        current_capacity += amount
        self.stored_energy = current_capacity
        return amount, current_capacity

    def keep_state(self, capacity: float) -> Any:
        """Keep state."""
        charging_delta = 0
        # after_capacity = capacity
        return charging_delta, capacity


class EVCharger(cp.Component):
    """Electric Vehicle Charger Component.

    Parameters
    ----------
        manufacturer : str
        name : str
        electric_vehicle : Vehicle_Pure
        sim_params : cp.SimulationParameters

    """

    # Inputs
    ElectricityInput = "ElectricityInput"
    # StoredEnergy = "StoredEnergy"
    # MaxStoredEnergy = "MaxStoredEnergy"
    EVChargerState = "EVChargerState"
    EVChargerMode = "EVChargerMode"
    MinimumStateOfCharge = "MinimumStateOfCharge"

    # Outputs
    StateOfCharge = "StateOfCharge"
    AfterStoredEnergy = "AfterStoredEnergy"
    Driving = "Driving"
    AtChargingStation = "AtChargingStation"
    ElectricityOutput = "ElectricityOutput"

    # Similar components to connect to:
    # 1. EVChargerController
    # 2. Some ChargingInput

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: EVChargerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(
            manufacturer=config.manufacturer,
            name=config.name,
            electric_vehicle=config.electric_vehicle,
            sim_params=my_simulation_parameters,
        )

        self.state = SimpleStorageState(
            max_var_val=self.charging_power,
            min_var_val=-self.charging_power,
            stored_energy=self.electric_vehicle.capacity,
            time_correction_factor=self.time_correction_factor,
            seconds_per_timestep=self.seconds_per_timestep,
        )

        self.previous_state = copy.deepcopy(self.state)

        self.charging_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )
        self.state_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.EVChargerState,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )
        self.mode_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.EVChargerMode,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )
        self.min_soc_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.MinimumStateOfCharge,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )

        self.soc_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.StateOfCharge, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.after_capacity_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.AfterStoredEnergy, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.driving_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.Driving, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.at_charging_station_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.AtChargingStation, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            sankey_flow_direction=True,
        )

    def build(
        self,
        manufacturer: str,
        name: str,
        electric_vehicle: Any,
        sim_params: SimulationParameters,
    ) -> None:
        """Build function."""
        self.time_correction_factor = 1 / sim_params.seconds_per_timestep
        self.seconds_per_timestep = sim_params.seconds_per_timestep

        electric_vehicle_charger_database = utils.load_smart_appliance("Wallbox")
        evcharger_found = False
        for evcharger in electric_vehicle_charger_database:
            if evcharger["Manufacturer"] == manufacturer and evcharger["Name"] == name:
                charging_power = evcharger["Charging Power"] * 1e3
                evcharger_found = True
                break

        if not evcharger_found:
            raise Exception("EV Charging Station has not been found in the database.")

        self.manufacturer = manufacturer
        self.name = name
        self.charging_power_original = charging_power * 1e-3
        self.charging_power = charging_power * self.time_correction_factor
        self.electric_vehicle = electric_vehicle

    def write_to_report(self) -> List[str]:
        """Write to resport."""
        lines = []
        lines.append(f"Name: {self.component_name}")
        lines.append(f"Manufacturer: {self.manufacturer}")
        lines.append(f"Model: {self.name}")
        lines.append(f"Charging Power: {self.charging_power_original} kW")
        lines.append(f"EV Battery Capacity: {self.electric_vehicle.max_capacity * 1e-3}")
        lines.append(f"Vehicle: {self.electric_vehicle.model}")
        return lines

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = copy.deepcopy(self.state)

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = copy.deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doubelchecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        if force_convergence:
            return

        # Gets inputs
        charging = stsv.get_input_value(self.charging_input_channel)
        state = stsv.get_input_value(self.state_channel)
        # mode = stsv.get_input_value(self.modeC)
        min_soc = stsv.get_input_value(self.min_soc_channel)

        capacity = self.state.stored_energy
        max_capacity = self.electric_vehicle.max_capacity

        min_capacity = max_capacity * min_soc

        if bool(self.electric_vehicle.car_in_charging_station[timestep]):
            connected_to_charging_station = 1
            driving = 0
            to_be_charged = -charging

            if state == 0:
                charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
            elif state == 1:
                charging_delta, after_capacity = self.charge_only_from_the_grid(
                    max_capacity=max_capacity, capacity=capacity
                )
            elif state == 2:
                charging_delta, after_capacity = self.charge_only_electricity_surplus(
                    to_be_charged=to_be_charged,
                    max_capacity=max_capacity,
                    capacity=capacity,
                )
            elif state == 3:
                charging_delta, after_capacity = self.operate_on_vehicle_to_grid(
                    to_be_charged=to_be_charged,
                    max_capacity=max_capacity,
                    min_capacity=min_capacity,
                    capacity=capacity,
                )
            else:
                raise Exception(
                    f"State {state} has not been implemented! " "Please check EVCharger and EVChargerController "
                )

        # Discharges battery according to electric vehicle use
        elif self.electric_vehicle.discharge[timestep] != 0.0:
            connected_to_charging_station = 0
            driving = 1

            charging_delta, after_capacity = self.state.withdraw(
                min_capacity=0,
                current_capacity=capacity,
                val=self.electric_vehicle.discharge[timestep],
            )
            charging_delta = 0
        else:
            connected_to_charging_station = 0
            driving = 0
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)

        stsv.set_output_value(self.soc_channel, after_capacity / max_capacity)
        stsv.set_output_value(self.after_capacity_channel, after_capacity)
        stsv.set_output_value(self.driving_channel, driving)
        stsv.set_output_value(self.electricity_output_channel, charging_delta)
        stsv.set_output_value(self.at_charging_station_channel, connected_to_charging_station)

    def charge_only_electricity_surplus(self, to_be_charged, max_capacity, capacity):
        """Charge only electricity surplus."""
        if to_be_charged >= 0:
            charging_delta, after_capacity = self.state.store(
                max_capacity=max_capacity, current_capacity=capacity, val=to_be_charged
            )
        else:
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
        return charging_delta, after_capacity

    def operate_on_vehicle_to_grid(
        self,
        to_be_charged: float,
        max_capacity: float,
        min_capacity: float,
        capacity: float,
    ) -> Any:
        """Operate o vehicle to grid."""
        if to_be_charged >= 0:
            charging_delta, after_capacity = self.state.store(
                max_capacity=max_capacity, current_capacity=capacity, val=to_be_charged
            )
        elif to_be_charged < 0:
            charging_delta, after_capacity = self.state.withdraw(
                min_capacity=min_capacity, current_capacity=capacity, val=to_be_charged
            )
        else:
            charging_delta, after_capacity = self.state.keep_state(capacity=capacity)
        return charging_delta, after_capacity

    def charge_only_from_the_grid(self, max_capacity, capacity):
        """Charge only from the grid."""
        charging_delta, after_capacity = self.state.force_store(max_capacity=max_capacity, current_capacity=capacity)
        return charging_delta, after_capacity


class EVChargerController(cp.Component):
    """Electricle vehicle controller class.

    Imports data from Load Profile Generator and
    uses as a Component Class in hisim.
    """

    # Inputs
    ElectricityInput = "ElectricityInput"
    StateOfCharge = "StateOfCharge"

    # Outputs
    EVChargerState = "EVChargerState"
    EVChargerMode = "EVChargerMode"
    MinimumStateOfCharge = "MinimumStateOfCharge"

    # Similar components to connect to:
    # 1. Some ChargingInput

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: EVChargerControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.mode = config.mode

        if self.mode == 1:
            self.mode_description = "Straight Charging"
            self.mode_extended_description = "Charge the Electric Vehicle whenever is connected to EV Charger"
        elif self.mode == 2:
            self.mode_description = "Charge only on Electricity Surplus"
            self.mode_extended_description = (
                "Charge the Electric Vehicle whenever the home grid is on electricity surplus"
            )
        elif self.mode == 3:
            self.mode_description = "Operate on Vehicle-to-Grid"
            self.mode_extended_description = (
                "Charge the Electric Vehicle whenever the home grid is on electricity surplus. "
                "Discharge the Electric Vehicle whenenver the home grid is running on electricity deficit."
            )
        elif self.mode == 4:
            self.mode_description = "Stepped Prioritized Charging"
            self.mode_extended_description = (
                "Charge only from the grid for SoC from 0% up to 60%. "
                "Charge only from the grid only on the electricity surplus for SoC from 60% up to 80%. "
                "Operate on Vehicle-to-Grid for SoC from 80% up to 100%"
            )
        elif self.mode == 5:
            self.mode_description = "Tight Stepped Prioritized Charging"
            self.mode_extended_description = (
                "Charge only from the grid for SoC from 0% up to 40%. "
                "Charge only from the grid only on the electricity surplus for SoC from 40% up to 70%. "
                "Operate on Vehicle-to-Grid for SoC from 70% up to 100%"
            )
        elif self.mode == 6:
            self.mode_description = "Super Tight Stepped Prioritized Charging"
            self.mode_extended_description = (
                "Charge only from the grid for SoC from 0% up to 20%. "
                "Charge only from the grid only on the electricity surplus for SoC from 20% up to 60%. "
                "Operate on Vehicle-to-Grid for SoC from 60% up to 100%"
            )
        else:
            self.mode_description = "WRITE MODE DESCRIPTION HERE!"
            self.mode_extended_description = "WRITE MODE EXTENDED DESCRIPTION HERE!"

        # Inputs
        self.charging_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )
        self.soc_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.StateOfCharge,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )

        self.state_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.EVChargerState, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.mode_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.EVChargerMode, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.min_soc_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.MinimumStateOfCharge,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
        )

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        # Gets inputs
        # charging = stsv.get_input_value(self.charging_inputC)
        state_of_charge = stsv.get_input_value(self.soc_channel)

        minimum_state_of_charge = 0.1

        # Straight Charging
        if self.mode == 1:
            state = 1
        # Charge only on Electricity Surplus
        elif self.mode == 2:
            state = 2
        # Operate on Vehicle-to-Grid
        elif self.mode == 3:
            state = 3
        # Stepped Prioritized Charging
        elif self.mode == 4:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.6:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif 0.6 <= state_of_charge < 0.8:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.8:
                minimum_state_of_charge = 0.8
                state = 3

        # Tight Stepped Prioritized Charging
        elif self.mode == 5:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.4:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif 0.4 <= state_of_charge < 0.7:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.7:
                minimum_state_of_charge = 0.7
                state = 3

        # Super Tight Stepped Prioritized Charging
        elif self.mode == 6:
            # Under this condition, charge only from the grid
            if state_of_charge < 0.2:
                state = 1
            # Under this condition, charge only on electricity surplus
            elif 0.2 <= state_of_charge < 0.6:
                state = 2
            # Under this condition, charge on Vehicle-to-Grid basis
            elif state_of_charge >= 0.6:
                minimum_state_of_charge = 0.6
                state = 3
        else:
            if self.mode is None:
                raise Exception("None mode is invalid.")
            raise Exception(f"Mode {self.mode} has not been implemented yet.")

        stsv.set_output_value(self.state_channel, state)
        stsv.set_output_value(self.min_soc_channel, minimum_state_of_charge)
        stsv.set_output_value(self.mode_channel, self.mode)

    def write_to_report(self):
        """Write to report."""
        lines = []
        lines.append(f"Mode Number: {self.mode}")
        lines.append(f"Mode: {self.mode_description}")
        try:
            lines.append(f"Extend Mode Description: {self.mode_extended_description}")
        except AttributeError:
            lines.append("Extend Mode Description: TO BE IMPLEMENTED")
        return lines



""" Controller for the generic_electrolyzer_h2 component. """

# clean
import os
from typing import List, Any
import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hisim.component import ConfigBase, Component, ComponentInput, ComponentOutput, SingleTimeStepValues, DisplayConfig

from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters
from hisim import log

__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "MIT"
__version__ = "0.5"
__maintainer__ = "Franz Oldopp"
__email__ = "f.oldopp@fz-juelich.de"
__status__ = "development"


@dataclass_json
@dataclass
class ElectrolyzerControllerConfig(ConfigBase):
    """Configutation of the Simple Electrolyzer Controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return ElectrolyzerController.get_full_classname()

    building_name: str
    name: str
    nom_load: float
    min_load: float
    max_load: float
    standby_load: float
    warm_start_time: float
    cold_start_time: float

    @classmethod
    def get_default_electrolyzer_controller_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get a default electrolyzer controller config."""
        config = ElectrolyzerControllerConfig(
            building_name=building_name,
            name="DefaultElectrolyzerController",
            nom_load=100.0,
            min_load=10.0,
            max_load=110.0,
            standby_load=5.0,
            warm_start_time=70.0,
            cold_start_time=1800.0,
        )
        return config

    @staticmethod
    def read_config(electrolyzer_name):
        """Opens the according JSON-file, based on the electrolyzer_name."""

        config_file = os.path.join(utils.HISIMPATH["inputs"], "electrolyzer_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data.get("Electrolyzer variants", {}).get(electrolyzer_name, {})

    @classmethod
    def control_electrolyzer(
        cls,
        electrolyzer_name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Initializes the config variables based on the JSON-file."""

        config_json = cls.read_config(electrolyzer_name)
        log.information("Electrolyzer config: " + str(config_json))

        config = ElectrolyzerControllerConfig(
            building_name=building_name,
            name="L1ElectrolyzerController",  # config_json.get("name", "")
            nom_load=config_json.get("nom_load", 0.0),
            min_load=config_json.get("min_load", 0.0),
            max_load=config_json.get("max_load", 0.0),
            standby_load=config_json.get("standby_load", 0.0),
            warm_start_time=config_json.get("warm_start_time", 0.0),
            cold_start_time=config_json.get("cold_start_time", 0.0),
        )
        return config


class ElectrolyzerController(Component):
    """Electrolyzer Controller class."""

    # Inputs
    ProvidedLoad = "ProvidedLoad"

    # Outputs
    DistributedLoad = "DistributedLoad"
    ShutdownCount = "ShutdownCount"
    StandbyCount = "StandbyCount"
    CurrentMode = "CurrentMode"
    CurtailedLoad = "CurtailedLoad"
    OffCount = "OffCount"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ElectrolyzerControllerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.controllerconfig = config

        self.nom_load = config.nom_load
        self.min_load = config.min_load
        self.max_load = config.max_load
        self.standby_load = config.standby_load
        self.warm_start_time = config.warm_start_time
        self.cold_start_time = config.cold_start_time

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels

        # Getting the load input
        self.load_input: ComponentInput = self.add_input(
            self.component_name,
            ElectrolyzerController.ProvidedLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            True,
        )

        # =================================================================================================================================
        # Output channels

        self.standby_count_total: ComponentOutput = self.add_output(
            self.component_name,
            ElectrolyzerController.StandbyCount,
            lt.LoadTypes.ON_OFF,
            lt.Units.ANY,
            output_description="standby count",
        )

        self.distributed_load: ComponentOutput = self.add_output(
            self.component_name,
            ElectrolyzerController.DistributedLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="Load to electrolyzer",
        )

        self.current_mode_electrolyzer: ComponentOutput = self.add_output(
            self.component_name,
            ElectrolyzerController.CurrentMode,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            output_description="current mode of electrolyzer",
        )

        self.curtailed_load: ComponentOutput = self.add_output(
            self.component_name,
            ElectrolyzerController.CurtailedLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="amount of curtailed load due to min and max load tresholds",
        )

        self.total_off_count: ComponentOutput = self.add_output(
            self.component_name,
            ElectrolyzerController.OffCount,
            lt.LoadTypes.ON_OFF,
            lt.Units.ANY,
            output_description="Total count of switching off",
        )
        # =================================================================================================================================
        # Initialize variables

        self.standby_count = 0.0
        self.current_state = "OFF"  # standby
        self.curtailed_load_count = 0.0
        self.off_count = 0.0
        self.activation_runtime = 0.0

        self.standby_count_previous = self.standby_count
        self.current_state_previous = self.current_state
        self.curtailed_load_count_previous = self.curtailed_load_count
        self.off_count_previous = self.off_count
        self.activation_runtime_previous = self.activation_runtime

    def load_check(self, current_load, min_load, max_load, standby_load):
        """Load check."""
        if None in (current_load, min_load, max_load, standby_load):
            raise ValueError(f"None type not accepted. {current_load}, {min_load}, {max_load}, {standby_load}")
        if current_load > max_load:
            current_load_to_system = max_load
            self.curtailed_load_count += current_load - max_load
            state = "ON"

        elif min_load <= current_load <= max_load:
            current_load_to_system = current_load
            self.curtailed_load_count += 0.0
            state = "ON"

        elif standby_load <= current_load < min_load:
            current_load_to_system = standby_load
            self.curtailed_load_count += current_load - standby_load
            state = "STANDBY"

        else:
            current_load_to_system = 0.0
            self.curtailed_load_count += current_load
            state = "OFF"

        return current_load_to_system, state, self.curtailed_load_count

    def state_check(self, target_state, cold_start_time_to_min, warm_start_time_to_min):
        """State check."""
        if target_state == "OFF":
            # System switches OFF
            if self.current_state == "ON":
                self.current_state = "SwitchingOFF"
                self.off_count += 1
            else:
                self.current_state = "OFF"

        elif target_state == "STANDBY":
            # System switches STANDY
            if self.current_state in ("OFF", "StartingfromOFF"):
                self.current_state = "OFF"
                self.off_count += 1
            if self.current_state == "ON":
                self.current_state = "SwitchingSTANDBY"
                self.standby_count += 1
            else:
                self.current_state = "STANDBY"

        else:
            # Test start
            if self.current_state in ["StartingfromOFF", "StartingfromSTANDBY"]:
                # pdb.set_trace()
                if self.activation_runtime <= self.my_simulation_parameters.seconds_per_timestep:
                    self.current_state = "Startingtomin"
                    # pdb.set_trace()
                else:
                    self.activation_runtime -= self.my_simulation_parameters.seconds_per_timestep
                    # pdb.set_trace()
                    # self.current_state = self.current_state
                    # starting to min auch unten aufnehmen um so min_load zu verteilen. "Starting to min" kann verwendet werden,
                    # da wenn wir wir durch die else: bedingungen da wieder raus kommen (theoretisch ;))

            # Test end
            elif self.current_state == "OFF":
                self.current_state = "StartingfromOFF"
                self.activation_runtime = cold_start_time_to_min
            elif self.current_state == "STANDBY":
                self.current_state = "StartingfromSTANDBY"
                self.activation_runtime = warm_start_time_to_min
            else:
                if self.activation_runtime > self.my_simulation_parameters.seconds_per_timestep:
                    self.activation_runtime -= self.my_simulation_parameters.seconds_per_timestep
                    # self.current_state = self.current_state
                # elif self.activation_runtime <= self.my_simulation_parameters.seconds_per_timestep:
                #    self.activation_runtime -= self.my_simulation_parameters.seconds_per_timestep
                #    self.current_state = self.current_state
                # elif self.activation_runtime <= 0.0:
                else:
                    self.activation_runtime = 0.0
                    self.current_state = "ON"

        return self.current_state, self.activation_runtime

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.standby_count_previous = self.standby_count
        self.current_state_previous = self.current_state
        self.curtailed_load_count_previous = self.curtailed_load_count
        self.off_count_previous = self.off_count
        self.activation_runtime_previous = self.activation_runtime

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.standby_count = self.standby_count_previous
        self.current_state = self.current_state_previous
        self.curtailed_load_count = self.curtailed_load_count_previous
        self.off_count = self.off_count_previous
        self.activation_runtime = self.activation_runtime_previous

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        if force_convergence:
            return
        """
        self.nom_load = config.nom_load
        self.min_load = config.min_load
        self.max_load = config.max_load
        self.warm_start_time = config.warm_start_time
        self.cold_start_time = config.cold_start_time
        """
        if self.nom_load == 0.0:
            self.nom_load = 1.0
        warm_start_time_to_min = self.warm_start_time * (self.min_load / self.nom_load)
        cold_start_time_to_min = self.cold_start_time * (self.min_load / self.nom_load)

        (current_load_to_system, state, self.curtailed_load_count) = self.load_check(
            (stsv.get_input_value(self.load_input)),
            self.min_load,
            self.max_load,
            self.standby_load,
        )  # change standby time
        # pdb.set_trace()
        (self.current_state, self.activation_runtime) = self.state_check(
            state, cold_start_time_to_min, warm_start_time_to_min
        )

        if self.current_state in ["OFF", "StartingfromOFF", "SwitchingOFF"]:
            stsv.set_output_value(self.distributed_load, 0.0)
            stsv.set_output_value(self.current_mode_electrolyzer, -1)
        elif self.current_state in [
            "STANDBY",
            "StartingfromSTANDBY",
            "SwitchingSTANDBY",
        ]:
            stsv.set_output_value(self.distributed_load, (self.nom_load * 0.05))
            stsv.set_output_value(self.current_mode_electrolyzer, 0)
        elif self.current_state == "Startingtomin":
            # pdb.set_trace()
            stsv.set_output_value(self.distributed_load, self.min_load)
            stsv.set_output_value(self.current_mode_electrolyzer, 1)

        else:
            stsv.set_output_value(self.distributed_load, current_load_to_system)
            stsv.set_output_value(self.current_mode_electrolyzer, 1)

        stsv.set_output_value(self.curtailed_load, self.curtailed_load_count)
        stsv.set_output_value(self.total_off_count, self.off_count)
        stsv.set_output_value(self.standby_count_total, self.standby_count)

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        lines = []
        for config_string in self.controllerconfig.get_string_dict():
            lines.append(config_string)
        lines.append("Component Name" + str(self.component_name))
        lines.append("Total curtailed load: " + str(self.curtailed_load_count) + " [kW]")
        lines.append("Number of times the system was switched off: " + str(self.off_count) + " [#]")
        lines.append("Number of times the system was switched to standby mode: " + str(self.standby_count) + " [#]")
        return lines
